"""This module implements the dantro PlotHelper class, which aims to abstract
matplotlib plot operations such that they can be made accessible for a
configuration-based declaration."""

import copy
import logging
import os
from collections import defaultdict
from itertools import product
from typing import Callable, Dict, Generator, Sequence, Tuple, Union

import numpy as np

from .._import_tools import LazyLoader
from ..exceptions import *
from ..tools import make_columns, recursive_update
from .utils.mpl import *

# Local constants and lazy loading modules that take a long time to import
log = logging.getLogger(__name__)

mpl = LazyLoader("matplotlib")
sns = LazyLoader("seaborn")
# NOTE matplotlib.pyplot is included on demand because lazy-loading is
#      somewhat unreliable for that module

# -----------------------------------------------------------------------------


class temporarily_changed_axis:
    """Context manager to temporarily change an axis in the
    :py:class:`.PlotHelper`.
    """

    def __init__(
        self, hlpr: "PlotHelper", tmp_ax_coords: Tuple[int, int] = None
    ):
        """Initialize the context manager.

        Args:
            hlpr (PlotHelper): The plot helper of which to select a temporary
                axis
            tmp_ax_coords (Tuple[int], optional): The coordinates of the
                temporary axis. If not given, will **not** change the axis.
        """
        self._hlpr = hlpr
        self._tmp_ax_coords = tmp_ax_coords
        self._old_ax_coords = None

    def __enter__(self):
        """Enter the context, selecting a temporary axis"""
        # Store the current axis' coordinates
        self._old_ax_coords = self._hlpr.ax_coords

        # If it needs to be changed, do it.
        if (
            self._tmp_ax_coords is not None
            and self._tmp_ax_coords != self._old_ax_coords
        ):
            log.debug(
                "Temporarily changing from axis %s to %s ...",
                self._old_ax_coords,
                self._tmp_ax_coords,
            )
            self._hlpr.select_axis(*self._tmp_ax_coords)

        else:
            log.trace("No need to change current axis.")

    def __exit__(self, *args):
        """Change back to the initial axis. Errors are not handled."""
        if self._old_ax_coords != self._hlpr.ax_coords:
            log.debug("Changing back to axis %s ...", self._old_ax_coords)
            self._hlpr.select_axis(*self._old_ax_coords)


def _coords_match(
    coords: Tuple[int], *, match: Union[tuple, str], full_shape: Tuple[int]
) -> bool:
    """Whether a coordinate is matched by a coordinate match tuple.

    Allowed values in the coordinate match tuple are:

        * integers: regarded as coordinates. If negative or exceeding the full
            shape, these are wrapped around.
        * Ellipsis: matches all coordinates
        * None: alias for Ellipsis

    Args:
        coords (Tuple[int]): The coordinate to match
        match (Union[tuple, str]): The match tuple, where None is
            interpreted as an Ellipsis and negative values are wrapped around
            by ``full_shape``. Can also be 'all', which is equivalent to a
            (None, None) tuple. Can also be a list, which is then converted to
            a tuple.
        full_shape (Tuple[int]): The full shape of the axes; needed to wrap
            around negative values in ``match``.

    Returns:
        bool: Whether ``coords`` matches ``match``

    Raises:
        TypeError: ``match`` not being a tuple or a list
        ValueError: Any of the arguments not being 2-tuples.
    """
    # Convert the 'all argument'
    match = match if match != "all" else (Ellipsis, Ellipsis)

    # Make sure it is a tuple, allowing conversion from lists
    if isinstance(match, list):
        match = tuple(match)

    elif not isinstance(match, tuple):
        raise TypeError(
            "Argument `match` needs to be a 2-tuple, list, or a "
            f"string, but was {type(match)} with value '{match}'!"
        )

    # Convert any Nones to Ellipsis
    match = tuple(m if m is not None else Ellipsis for m in match)

    # Check length and values, not allowing values exceeding the shape
    if any([len(t) != 2 for t in (coords, match, full_shape)]):
        raise ValueError(
            "Need 2-tuples for arguments, got {}, {}, and {}!".format(
                coords, match, full_shape
            )
        )

    elif not all([m is Ellipsis or m < s for m, s in zip(match, full_shape)]):
        raise ValueError(
            f"Got match values {match} exceeding the shape {full_shape}! Take "
            "care that all values are strictly smaller than the "
            "maximum value. Negative values are allowed and will "
            "be evaluated via a modulo operation."
        )

    for c, m, s in zip(coords, match, full_shape):
        if m is Ellipsis:
            # Always matches
            continue
        if (m % s) != c:
            # No match
            return False

    # Went through -> have a match
    return True


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------


class PlotHelper:
    """The PlotHelper takes care of the figure setup and saving and allows
    accessing matplotlib utilities through the plot configuration.
    """

    _SPECIAL_CFG_KEYS: Tuple[str] = ("setup_figure", "save_figure")
    """Configuration keys with special meaning"""

    _FIGURE_HELPERS: Tuple[str] = (
        "align_labels",
        "set_suptitle",
        "set_figlegend",
        "subplots_adjust",
        "figcall",
    )
    """Names of those helpers that are applied on the figure level"""

    def __init__(
        self,
        *,
        out_path: str,
        helper_defaults: dict = None,
        update_helper_cfg: dict = None,
        raise_on_error: bool = True,
        animation_enabled: bool = False,
    ):
        """Initialize a PlotHelper with a certain configuration.

        This configuration is the so-called "base" configuration and is not
        axis-specific. There is the possibility to specify axis-specific
        configuration entries.

        All entries in the helper configuration are deemed 'enabled' unless
        they explicitly specify ``enabled: false`` in their configuration.

        Args:
            out_path (str): path to store the created figure. This may be an
                absolute path or a relative path; the latter is regarded as
                relative to the current working directory. The home directory
                indicator ``~`` is expanded.
            helper_defaults (dict, optional): The basic configuration of the
                helpers.
            update_helper_cfg (dict, optional): A configuration used to update
                the existing helper defaults
            raise_on_error (bool, optional): Whether to raise on an exception
                created on helper invocation or just log the error
            animation_enabled (bool, optional): Whether animation mode is
                enabled.
        """
        # Determine available helper methods, store it as tuple
        self._AVAILABLE_HELPERS = [
            attr_name[6:]
            for attr_name in dir(self)
            if attr_name.startswith("_hlpr_")
            and callable(getattr(self, attr_name))
        ]
        self._AVAILABLE_HELPERS = tuple(self._AVAILABLE_HELPERS)

        # Store (a copy of) the base configuration
        self._base_cfg = (
            copy.deepcopy(helper_defaults) if helper_defaults else {}
        )

        # Update the defaults
        if update_helper_cfg:
            self._base_cfg = recursive_update(
                self._base_cfg, copy.deepcopy(update_helper_cfg)
            )

        # Extract the axis-specific update list
        self._axis_specific_updates = self._base_cfg.pop("axis_specific", {})

        # Check that all remaining entries are valid keys
        self._raise_on_invalid_helper_name(
            *self._base_cfg.keys(), special_cfg_keys_allowed=True
        )

        # Initialize the actual axis-specific configuration empty; it can only
        # be compiled once the figure is created.
        self._cfg = None

        # Store the other attributes
        self._out_path = os.path.expanduser(out_path)
        self._raise_on_error = raise_on_error
        self._animation_enabled = animation_enabled

        # Initialize attributes that are set at a later point
        self._fig = None
        self._axes = None
        self._current_ax_coords = None

        self._animation_update = None
        self._invoke_before_grab = False

        # Storage of figure-level and axis-level objects or attributes
        self._additional_axes = None
        self._handles_labels = defaultdict(dict)
        self._figlegend = None
        self._attrs = dict()
        self._ax_attrs = dict()

        log.debug("PlotHelper initialized.")

    # .. Properties ...........................................................

    @property
    def _axis_cfg(self) -> dict:
        """Return the configuration for the current axis; not a deep copy!"""
        try:
            return self._cfg[self.ax_coords]

        except RuntimeError as err:
            raise RuntimeError(
                "The axis-specific configuration is only available while a "
                "figure is associated with the PlotHelper!"
            ) from err

    @property
    def axis_cfg(self) -> dict:
        """Returns a deepcopy of the current axis' configuration."""
        return copy.deepcopy(self._axis_cfg)

    @property
    def base_cfg(self) -> dict:
        """Returns a deepcopy of the base configuration, i.e. the configuration
        that is *not* axis-specific.
        """
        return copy.deepcopy(self._base_cfg)

    @property
    def fig(self) -> "matplotlib.figure.Figure":
        """Returns the current figure"""
        if self._fig is None:
            raise ValueError(
                "No figure initialized or already closed! Use "
                "the `setup_figure` method to create a figure "
                "instance."
            )
        return self._fig

    @property
    def ax(self) -> "matplotlib.axis.Axis":
        """Returns the current axis of the associated figure"""
        return self.fig.gca()

    @property
    def ax_coords(self) -> Tuple[int]:
        """Returns the current axis coordinates within a subfigure in shape
        ``(col, row)``.

        For example, the ``(0, 0)`` coordinate refers to the top left subplot
        of the figure. ``(1, 2)`` is the axis object in the second column,
        third row.
        """
        if self._current_ax_coords is None:
            raise RuntimeError(
                "The current axis coordinate is only defined "
                "while a figure is associated with the "
                "PlotHelper!"
            )

        return self._current_ax_coords
        # NOTE There _would_ be the possiblity to use the matplotlib axis
        #      properties self.ax.numCol and .numRow, which store the column
        #      and row the axis was created in via plt.subplots. However, that
        #      that information gets lost as soon as an additional axis is
        #      added to the figure, e.g. when adding a color bar. Thus, the
        #      current axis coordinates need to be stored in an attribute.

    @property
    def axes(self) -> np.ndarray:
        """Returns the axes array, which is of shape ``(#cols, #rows)``.

        The (0, 0) axis refers to the top left subplot of the figure.
        """
        return self._axes

    @property
    def available_helpers(self) -> Tuple[str]:
        """List of available helper names"""
        return self._AVAILABLE_HELPERS

    @property
    def enabled_helpers(self) -> list:
        """Returns a list of enabled helpers *for the current axis*"""
        return [
            helper_name
            for helper_name in self._axis_cfg
            if self._axis_cfg[helper_name].get("enabled", True)
            and helper_name
            not in (self._SPECIAL_CFG_KEYS + self._FIGURE_HELPERS)
        ]

    @property
    def enabled_figure_helpers(self) -> list:
        """Returns a list of enabled figure-level helpers"""
        return [
            helper_name
            for helper_name in self._base_cfg
            if self._base_cfg[helper_name].get("enabled", True)
            and helper_name in self._FIGURE_HELPERS
        ]

    @property
    def out_path(self) -> str:
        """Returns the output path of the plot"""
        return self._out_path

    @property
    def animation_enabled(self) -> bool:
        """Whether animation mode is currently enabled or not"""
        return self._animation_enabled

    @property
    def animation_update(self) -> Callable:
        """Returns the animation update generator callable"""
        if self._animation_update is None:
            raise ValueError(
                "No animation update generator was registered "
                "with the PlotHelper! Cannot perform animation "
                "update."
            )
        return self._animation_update

    @property
    def invoke_before_grab(self) -> bool:
        """Whether the helpers are to be invoked before grabbing each frame of
        an animation.
        """
        return self._invoke_before_grab

    @property
    def raise_on_error(self) -> bool:
        """Whether the PlotHelper was configured to raise exceptions"""
        return self._raise_on_error

    @property
    def axis_handles_labels(self) -> Tuple[list, list]:
        """Returns the tracked axis handles and labels for the current axis"""
        return (
            self._handles_labels[self.ax_coords].get("handles", []),
            self._handles_labels[self.ax_coords].get("labels", []),
        )

    @property
    def all_handles_labels(self) -> Tuple[list, list]:
        """Returns all associated handles and labels"""
        handles, labels = [], []
        for hl in self._handles_labels.values():
            handles += hl.get("handles", [])
            labels += hl.get("labels", [])

        return handles, labels

    @property
    def axis_is_empty(self) -> bool:
        """Returns true if the current axis is empty, i.e. has no artists added
        to it: Basically, negation of :py:meth:`matplotlib.axes.Axes.has_data`.
        """
        return not self.ax.has_data()

    # .. Figure setup and axis control ........................................

    def attach_figure_and_axes(
        self, *, fig, axes, skip_if_identical: bool = False
    ) -> None:
        """Attaches the given figure and axes to the PlotHelper. This method
        replaces an existing figure and existing axes with the ones given.

        As the PlotHelper relies on axes being accessible via coordinate pairs,
        multiple axes must be passed as two-dimensional array-like. Since the
        axes are internally stored as numpy array, the axes-grid must be
        complete.

        Note that by closing the old figure the existing axis-specific config
        and all existing axes are destroyed. In other words: All information
        previously provided via the provide_defaults and the ``mark_*`` methods
        is lost. Therefore, if needed, it is recommended to call this method at
        the beginning of the plotting function.

        .. note::

            This function assumes multiple axes to be passed in (y,x) format
            (as e.g. returned by matplotlib.pyplot.subplots with squeeze set to
            False) and internally transposes the axes-grid such that afterwards
            it is accessible via (x,y) coordinates.

        Args:
            fig: The new figure which replaces the existing.
            axes: single axis or 2d array-like containing the axes
            skip_if_identical (bool, optional): If True, will check if the
                given ``fig`` is *identical* to the already associated figure;
                if so, will do nothing. This can be useful if one cannot be
                sure if the figure was already associated. In such a case, note
                that the ``axes`` argument is completely ignored.

        Raises:
            ValueError: On multiple axes not being passed in 2d format.

        Returns:
            None

        """
        if skip_if_identical and fig is self._fig:
            log.debug("Figure was already associated; not doing anything.")
            return

        # Prepare the new axis object
        try:
            # Assuming it's a scalar, np.reshape leads to np.array being called
            # on the object, thus allowing any scalar type. Only in cases where
            # an np.ndarray with size > 1 is given will this reshape operation
            # fail.
            axes = np.reshape(axes, (1, 1))

        except ValueError:
            # Else, assume array-like, containing the axes in (y,x) format.
            # Transpose the axes such that they are accessible in the (x,y)
            # format, which is used internally throughout the PlotHelper.
            axes = np.array(axes).T

        # Ensure correct shape
        if axes.ndim != 2:
            raise ValueError(
                "When attaching a figure with multiple axes, the "
                "axes must be passed as a 2d array-like object! "
                f"Got object of shape {axes.shape}."
            )

        # Everything ok, close figure and attach new figure and axes
        log.debug(
            "Closing existing figure and re-associating with a new figure ..."
        )
        self.close_figure()

        self._fig = fig
        self._axes = axes

        log.debug("Figure %d and axes attached.", fig.number)

        # Reset some tracking attributes
        self._handles_labels = defaultdict(dict)
        self._additional_axes = None
        self._figlegend = None

        # Can now evaluate the axis-specific configuration
        self._cfg = self._compile_axis_specific_cfg()

        # Select the (0, 0) axis, for consistency
        self.select_axis(0, 0)

    def setup_figure(self, **update_fig_kwargs):
        """Sets up a matplotlib figure instance and axes with the given
        configuration (by calling :py:func:`matplotlib.pyplot.subplots`) and
        attaches both to the PlotHelper.

        If the ``scale_figsize_with_subplots_shape`` option is enabled here,
        this method will also take care of scaling the figure accordingly.

        Args:
            **update_fig_kwargs: Parameters that are used to update the
                figure setup parameters stored in ``setup_figure``.
        """
        import matplotlib.pyplot as plt

        fig_kwargs = self.base_cfg.get("setup_figure", {})

        if update_fig_kwargs:
            fig_kwargs = recursive_update(fig_kwargs, update_fig_kwargs)

        # Need to handle scaling argument separately
        scale_figsize = fig_kwargs.pop(
            "scale_figsize_with_subplots_shape", False
        )

        # Now, create the figure and axes and attach them
        fig, axes = plt.subplots(squeeze=False, **fig_kwargs)
        log.debug("Figure %d created.", fig.number)

        self.attach_figure_and_axes(fig=fig, axes=axes)

        # Scale figure, if needed
        if scale_figsize and self.axes.size > 1:
            log.debug(
                "Scaling current figure size with subplots shape %s ...",
                self.axes.shape,
            )

            old_figsize = self.fig.get_size_inches()
            self.fig.set_size_inches(
                old_figsize[0] * self.axes.shape[0],
                old_figsize[1] * self.axes.shape[1],
            )

            log.debug(
                "Scaled figure size from %s to %s.",
                old_figsize,
                self.fig.get_size_inches(),
            )

    def save_figure(self, *, close: bool = True):
        """Saves and (optionally, but default) closes the current figure

        Args:
            close (bool, optional): Whether to close the figure after saving.
        """
        self.fig.savefig(self.out_path, **self.base_cfg.get("save_figure", {}))
        log.debug("Figure saved at: %s", self.out_path)

        if close:
            self.close_figure()

    def close_figure(self):
        """Closes the figure and disassociates it from the helper. This method
        has no effect if no figure is currently associated.

        This also removes the axes objects and deletes the axis-specific
        configuration. All information provided via provide_defaults and the
        ``mark_*`` methods is lost.
        """
        import matplotlib.pyplot as plt

        if self._fig is None:
            log.debug("No figure currently associated; nothing to close.")
            return

        fignum = self.fig.number
        plt.close(self.fig)
        log.debug("Figure %d closed.", fignum)

        self._fig = None
        self._axes = None
        self._current_ax_coords = None
        self._cfg = None
        self._additional_axes = None
        self._handles_labels = defaultdict(list)
        self._figlegend = None
        log.debug("Associated data removed.")

    def select_axis(self, col: int = None, row: int = None, *, ax=None):
        """Sets the current axis.

        Setting the axis can happen in three ways, depending on the arguments:

            - The axis object at the given ``col`` and ``row`` coordinates.
            - An explicitly given axis object (if ``ax`` is given)
            - The current axis (if all arguments are None)

        This method can be used to change to a different associated axis to
        continue plotting on that axis.

        Calling this method may also become necessary if the current axis is
        changed in a part of the program where the plot helper is not
        involved; in such a case, the currently selected axis may have been
        changed directly via the matplotlib interface.
        This method can then be used to synchronize the two again.

        .. note::

            The ``col`` and ``row`` values are wrapped around according to the
            shape of the associated ``axes`` array, thereby allowing to specify
            them as negative values for indexing from the back.

        Args:
            col (int, optional): The column to select, i.e. the x-coordinate.
                Can be negative, in which case it indexes backwards from the
                last column.
            row (int, optional): The row to select, i.e. the y-coordinate. Can
                be negative, in which case it indexes backwards from the last
                row.
            ax (optional): If given this axis object, tries to look it up from
                the associated axes array.

        Raises:
            ValueError: On failing to set the current axis or if the given axis
                object or the result of :py:func:`matplotlib.pyplot.gca` was
                not part of the associated axes array.
                To associate the correct figure and axes, use the
                :py:meth:`.attach_figure_and_axes` method.

        """
        import matplotlib.pyplot as plt

        # Column and row need either both be given or both NOT be given. (XOR)
        if (col is None) != (row is None):
            raise ValueError(
                "Need both `col` and `row` arguments to select the current "
                f"axis via its coordinates! Got  col: {col},  row: {row}"
            )

        # Without any arguments, use the current axis for the lookup
        if col is None and row is None and ax is None:
            ax = plt.gca()

        # With an axis argument now given, it means that we need to retrieve
        # the col and row values via the axis object
        if ax is not None:
            # ... and we require that `col` and `row` are not given.
            if col is not None or row is not None:
                raise ValueError(
                    "Cannot specify arguments `col` and/or `row` if also "
                    "setting the `ax` argument for `select_axis`!"
                )

            # Look up column and row
            col, row = self._find_axis_coords(ax)

        # Wrap around negative values
        col = col if col >= 0 else col % self.axes.shape[0]
        row = row if row >= 0 else row % self.axes.shape[1]

        # Now ready to select
        log.debug("Selecting axis (%d, %d) ...", col, row)

        try:
            self.fig.sca(self.axes[col, row])

        except IndexError as exc:
            raise ValueError(
                f"Could not select axis ({col}, {row}) from figure with "
                f"associated axes array of shape {self.axes.shape}!"
            ) from exc

        else:
            self._current_ax_coords = (
                col % self.axes.shape[0],
                row % self.axes.shape[1],
            )

            log.debug(
                "Axis (%d, %d) selected. Enabled helpers: %s",
                self.ax_coords[0],
                self.ax_coords[1],
                ", ".join(self.enabled_helpers),
            )

    def coords_iter(
        self,
        *,
        match: Union[tuple, str] = None,
    ) -> Generator[Tuple[int], None, None]:
        """Returns a generator to iterate over all coordinates that match
        ``match``.

        Args:
            match (Union[tuple, str]): The coordinates to match; those that do
                not match this pattern (evaluated by :py:func:`._coords_match`)
                will not be yielded. If not given, will iterate only over the
                currently selected axis.

        Yields:
            Generator[Tuple[int], None, None]: The axis coordinates generator
        """
        # If no match argument is given, match only current coordinates
        match = match if match is not None else self.ax_coords

        # Go over the cartesian product ranges specified by the subplots shape
        for coords in product(
            range(self.axes.shape[0]), range(self.axes.shape[1])
        ):
            if _coords_match(coords, match=match, full_shape=self.axes.shape):
                yield coords

    # .........................................................................
    # Helper invocation and configuration

    def _invoke_helper(
        self,
        helper_name: str,
        *,
        axes: Union[tuple, str] = None,
        mark_disabled_after_use: bool = True,
        raise_on_error: bool = None,
        **update_kwargs,
    ) -> None:
        """Invokes a single helper on the specified axes, if it is enabled, and
        marks it disabled afterwards. The given update parameters are used to
        update the existing configuration.

        Unlike the public invoke_helper method, this method checks for whether
        the helper is enabled, while the public method automatically assumes
        it is meant to be enabled.

        Args:
            helper_name (str): helper which is invoked
            axes (Union[tuple, str], optional): A coordinate match tuple of
                the axes to invoke this helper on. If not given, will invoke
                only on the current axes.
            mark_disabled_after_use (bool, optional): If True, the helper is
                marked as disabled after invoking it
            raise_on_error (bool, optional): If given, overwrites the default
                controlled via the ``raise_on_error`` attribute.
            **update_kwargs: Update parameters for this specific plot helper.
                Note that these do not persist, but are only used for this
                invocation.

        Raises:
            PlotHelperErrors: On failing plot helper invocations
            ValueError: No matching helper function defined
        """
        is_figure_helper = helper_name in self._FIGURE_HELPERS

        def invoke_now(
            helper_name: str, errors: list, *, ax_coords: tuple = None
        ):
            """Invokes a single helper and keeps track of errors by mutably
            appending to the given ``errors`` list.
            """
            # Prepare the helper params, always working on a copy.
            # Depending on the helper, these can be either axis-specific or
            # figure-specific configurations.
            if not is_figure_helper:
                helper_params = self.axis_cfg.get(helper_name, {})
            else:
                helper_params = self.base_cfg.get(helper_name, {})

            if update_kwargs:
                recursive_update(helper_params, copy.deepcopy(update_kwargs))

            # If it was disabled, have nothing else to do here
            if not helper_params.pop("enabled", True):
                log.debug(
                    "Helper '%s' is not enabled for axis %s.",
                    helper_name,
                    self.ax_coords,
                )
                return

            # Also skip it if it is an axis-level helper and the axis is empty
            if (
                not is_figure_helper
                and helper_params.pop("skip_empty_axes", True)
                and self.axis_is_empty
            ):
                log.debug(
                    "Helper '%s' skipped for empty axis %s.",
                    helper_name,
                    self.ax_coords,
                )
                return

            # Invoke helper
            log.debug(
                "Invoking helper function '%s' on axis %s ...",
                helper_name,
                self.ax_coords,
            )

            try:
                helper(**helper_params)

            except Exception as exc:
                errors.append(
                    PlotHelperError(
                        exc,
                        name=helper_name,
                        ax_coords=ax_coords,
                        params=helper_params,
                    )
                )

            if mark_disabled_after_use:
                self.mark_disabled(helper_name)

        # Initial checks
        try:
            helper = getattr(self, "_hlpr_" + helper_name)

        except AttributeError as err:
            _available = ", ".join(self.available_helpers)
            raise PlotConfigError(
                f"No helper with name '{helper_name}' available! "
                f"Available helpers: {_available}"
            ) from err

        # A list to collect error messages in; is updated mutably.
        errors = []

        # If this is a figure-level helper, handle separately:
        if is_figure_helper:
            invoke_now(helper_name, errors)
            self._handle_errors(*errors, raise_on_error=raise_on_error)
            return

        # Go over all matching axis coordinates, temporarily change to an axis
        # and invoke the helper there. Keeps track of
        for ax_coords in self.coords_iter(match=axes):
            with temporarily_changed_axis(self, tmp_ax_coords=ax_coords):
                invoke_now(helper_name, errors)

        # Finally, handle the gathered errors
        self._handle_errors(*errors, raise_on_error=raise_on_error)

    def invoke_helper(
        self,
        helper_name: str,
        *,
        axes: Union[tuple, str] = None,
        mark_disabled_after_use: bool = True,
        raise_on_error: bool = None,
        **update_kwargs,
    ):
        """Invokes a single helper on the specified axes.

        Args:
            helper_name (str): The name of the helper to invoke
            axes (Union[tuple, str], optional): A coordinate match tuple of
                the axes to invoke this helper on. If not given, will invoke
                only on the current axes.
            mark_disabled_after_use (bool, optional): If True, the helper is
                marked as disabled after the invocation.
            raise_on_error (bool, optional): If given, overwrites the default
                controlled via the ``raise_on_error`` attribute.
            **update_kwargs: Update parameters for this specific plot helper.
                Note that these do not persist, but are only used for this
                invocation.

        Raises:
            PlotHelperErrors: On failing plot helper invocation
        """
        self._invoke_helper(
            helper_name,
            axes=axes,
            enabled=True,
            mark_disabled_after_use=mark_disabled_after_use,
            raise_on_error=raise_on_error,
            **update_kwargs,
        )

    def invoke_helpers(
        self,
        *helper_names,
        axes: Union[tuple, str] = None,
        mark_disabled_after_use: bool = True,
        raise_on_error: bool = None,
        **update_helpers,
    ):
        """Invoke all specified helpers on the specified axes.

        Args:
            *helper_names: The helper names to invoke
            axes (Union[tuple, str], optional): A coordinate match tuple of
                the axes to invoke this helper on. If not given, will invoke
                only on the current axes.
            mark_disabled_after_use (bool, optional): Whether to mark helpers
                disabled after they were used.
            raise_on_error (bool, optional): If given, overwrites the default
                controlled via the ``raise_on_error`` attribute.
            **update_helpers: Update parameters for all plot helpers.
                These have to be grouped under the name of the helper in order
                to be correctly associated. Note that these do not persist,
                but are only used for this invocation.

        Raises:
            PlotHelperErrors: On failing plot helper invocations
        """
        errors = []
        for helper_name in helper_names:
            try:
                self.invoke_helper(
                    helper_name,
                    axes=axes,
                    mark_disabled_after_use=mark_disabled_after_use,
                    raise_on_error=True,
                    **update_helpers.get(helper_name, {}),
                )

            except PlotHelperErrors as pherrs:
                errors += pherrs.errors

        self._handle_errors(*errors, raise_on_error=raise_on_error)

    def invoke_enabled(
        self,
        *,
        axes: Union[tuple, str] = None,
        mark_disabled_after_use: bool = True,
        raise_on_error: bool = None,
        **update_helpers,
    ):
        """Invokes all enabled helpers with their current configuration on the
        matching axes and all enabled figure-level helpers on the figure.

        Internally, this first invokes all figure-level helpers and then calls
        :py:meth:`~dantro.plot.plot_helper.PlotHelper.invoke_helpers`
        with all enabled helpers for all axes matching the ``axes`` argument.

        .. note::

            When setting ``mark_disabled_after_use = False``, this will lead to
            figure-level helpers being invoked multiple times. As some of these
            helpers do not allow multiple invocation, invoking this method a
            second time *might* fail if not disabling them as part of the first
            call to this method.

        Args:
            axes (Union[tuple, str], optional): A coordinate match tuple of
                the axes to invoke this helper on. If not given, will invoke
                only on the current axes.
            mark_disabled_after_use (bool, optional): If True, the helper is
                marked as disabled after the invocation.
            raise_on_error (bool, optional): If given, overwrites the default
                controlled via the ``raise_on_error`` attribute.
            **update_helpers: Update parameters for all plot helpers.
                These have to be grouped under the name of the helper in order
                to be correctly associated. Note that these do not persist,
                but are only used for this invocation.

        Raises:
            PlotHelperErrors: On failing plot helper invocations
        """

        def invoke(*helper_names, errors: list):
            """Small helper to invoke the given helpers and keep track of
            errors via the given ``errors`` list (mutably changed)
            """
            try:
                self.invoke_helpers(
                    *helper_names,
                    mark_disabled_after_use=mark_disabled_after_use,
                    raise_on_error=True,
                    **update_helpers,
                )

            except PlotHelperErrors as pherrs:
                errors += pherrs.errors

        # Mutably extended list of errors
        errors = []

        # Handle figure-level helpers first ...
        invoke(*self.enabled_figure_helpers, errors=errors)

        # ... and then axis-level helpers, each on their own axis
        for ax_coords in self.coords_iter(match=axes):
            with temporarily_changed_axis(self, tmp_ax_coords=ax_coords):
                invoke(*self.enabled_helpers, errors=errors)

        self._handle_errors(*errors, raise_on_error=raise_on_error)

    def provide_defaults(
        self,
        helper_name: str,
        *,
        axes: Union[tuple, str] = None,
        mark_enabled: bool = True,
        **update_kwargs,
    ):
        """Update or add a single entry to a helper's configuration.

        Args:
            helper_name (str): The name of the helper whose config is to change
            axes (Union[tuple, str], optional): A coordinate match tuple of
                the axes to invoke this helper on. If not given, will invoke
                only on the current axes.
            mark_enabled (bool, optional): Whether to mark the helper enabled
                by providing defaults
            **update_kwargs: dict containing the helper parameters with
                which the config is updated recursively
        """
        self._raise_on_invalid_helper_name(
            helper_name, special_cfg_keys_allowed=True
        )

        # Special configuration keys and figure-level helpers are updating the
        # base configuration
        if helper_name in (self._SPECIAL_CFG_KEYS + self._FIGURE_HELPERS):
            if helper_name not in self._base_cfg:
                self._base_cfg[helper_name] = dict()

            # Do an in-place recursive update; can return afterwards
            recursive_update(
                self._base_cfg[helper_name], copy.deepcopy(update_kwargs)
            )
            log.debug(
                "Updated defaults for special configuration entry '%s'.",
                helper_name,
            )
            return

        # Go over all selected axes
        for ax_coords in self.coords_iter(match=axes):
            # Update or create the configuration entry, manually passing the
            # coordinates. Using the _axis_cfg property would require to enter
            # the temporary axis context, which is overkill at this point.
            if helper_name not in self._cfg[ax_coords]:
                # Create a new empty entry
                self._cfg[ax_coords][helper_name] = dict()

            # Can do an in-place recursive update now
            recursive_update(
                self._cfg[ax_coords][helper_name], copy.deepcopy(update_kwargs)
            )
            log.debug(
                "Updated axis-specific defaults for helper '%s' and axis %s.",
                helper_name,
                ax_coords,
            )

            if mark_enabled and helper_name not in self._SPECIAL_CFG_KEYS:
                self.mark_enabled(helper_name, axes=ax_coords)

    def mark_enabled(self, *helper_names, axes: Union[tuple, str] = None):
        """Marks the specified helpers as enabled for the specified axes.

        Args:
            *helper_names: Helpers to be enabled
            axes (Union[tuple, str], optional): A coordinate match tuple of
                the axes to invoke this helper on. If not given, will invoke
                only on the current axes.
        """
        # Check the helper names
        self._raise_on_invalid_helper_name(*helper_names)

        # Handle the figure-level helpers separately
        fig_helpers = [hn for hn in helper_names if hn in self._FIGURE_HELPERS]
        helper_names = [hn for hn in helper_names if hn not in fig_helpers]

        for helper_name in fig_helpers:
            if helper_name not in self._base_cfg:
                self._base_cfg[helper_name] = dict()

            self._base_cfg[helper_name]["enabled"] = True
            log.debug(
                "Marked figure-level helper '%s' as enabled.", helper_name
            )

        # Go over all selected axes
        for ax_coords in self.coords_iter(match=axes):
            # Go over the specified helper names
            for helper_name in helper_names:
                # Create the empty dict, if necessary
                if helper_name not in self._cfg[ax_coords]:
                    self._cfg[ax_coords][helper_name] = dict()

                self._cfg[ax_coords][helper_name]["enabled"] = True
                log.debug(
                    "Marked helper '%s' enabled for axis %s.",
                    helper_name,
                    ax_coords,
                )

    def mark_disabled(self, *helper_names, axes: Union[tuple, str] = None):
        """Marks the specified helpers as disabled for the specified axes.

        Args:
            *helper_names: Helpers to be disabled
            axes (Union[tuple, str], optional): A coordinate match tuple of
                the axes to invoke this helper on. If not given, will invoke
                only on the current axes.
        """
        # Check the helper names
        self._raise_on_invalid_helper_name(*helper_names)

        # Handle the figure-level helpers separately
        fig_helpers = [hn for hn in helper_names if hn in self._FIGURE_HELPERS]
        helper_names = [hn for hn in helper_names if hn not in fig_helpers]

        for helper_name in fig_helpers:
            if helper_name not in self._base_cfg:
                self._base_cfg[helper_name] = dict()

            self._base_cfg[helper_name]["enabled"] = False
            log.debug(
                "Marked figure-level helper '%s' as disabled.", helper_name
            )

        # Go over all selected axes
        for ax_coords in self.coords_iter(match=axes):
            # Go over the specified helper names
            for helper_name in helper_names:
                # Create the empty dict, if necessary
                if helper_name not in self._cfg[ax_coords]:
                    self._cfg[ax_coords][helper_name] = dict()

                self._cfg[ax_coords][helper_name]["enabled"] = False
                log.debug(
                    "Marked helper '%s' disabled for axis %s.",
                    helper_name,
                    ax_coords,
                )

    # .........................................................................
    # Methods to track artists (e.g. legend handles and labels)

    def track_handles_labels(self, handles: list, labels):
        """Keep track of legend handles and/or labels for the current axis.

        Args:
            handles (list): The handles to keep track of
            labels (list): The associated labels
        """
        if len(handles) != len(labels):
            raise ValueError(
                "Lists of handles and labels need to be of the same size but "
                f"were {len(handles)} != {len(labels)}, respectively."
            )

        if self.ax_coords not in self._handles_labels:
            self._handles_labels[self.ax_coords] = defaultdict(list)

        self._handles_labels[self.ax_coords]["handles"] += handles
        self._handles_labels[self.ax_coords]["labels"] += labels

        log.debug(
            "Axis %s: now tracking %d handles and labels (%d new).",
            self.ax_coords,
            len(self._handles_labels[self.ax_coords]["handles"]),
            len(handles),
        )

    # .........................................................................
    # Animation interface

    def register_animation_update(
        self,
        update_func: Callable,
        *,
        invoke_helpers_before_grab: bool = False,
    ):
        """Registers a generator used for animations.

        Args:
            update_func (Callable): Generator object over which is iterated
                over to create an animation. This needs
            invoke_helpers_before_grab (bool, optional): Whether to invoke all
                enabled plot helpers before grabbing a frame. This should be
                set to ``True`` if the animation update function overwrites the
                effects of the previously applied helpers.
        """
        self._animation_update = update_func
        self._invoke_before_grab = invoke_helpers_before_grab
        log.debug("Registered animation update function.")

    def enable_animation(self):
        """Can be invoked to enter animation mode. An action is only performed
        if the helper is **not** currently in animation mode.

        Raises:
            EnterAnimationMode: Conveys to the plot creator that animation
                mode is to be entered.
        """
        if not self.animation_enabled:
            log.debug("Enabling animation mode ...")
            self._animation_enabled = True
            raise EnterAnimationMode()
        log.debug("Animation mode is already enabled.")

    def disable_animation(self):
        """Can be invoked to exit animation mode. An action is only performed
        if the helper **is** currently in animation mode.

        Raises:
            ExitAnimationMode: Conveys to the plot creator that animation mode
                should be left.
        """
        if self.animation_enabled:
            log.debug("Disabling animation mode ...")
            self._animation_enabled = False
            raise ExitAnimationMode()
        log.debug("Animation mode is already disabled.")

    # .........................................................................
    # Private support methods

    def _compile_axis_specific_cfg(self) -> Dict[tuple, dict]:
        """With a figure set up, compiles the axis-specific helper config."""
        # The target dict of axis coordinates to fully resolved dicts
        cfg = dict()

        # Compile a base configuration dict that does not contain special keys
        base_cfg = {
            k: v
            for k, v in self._base_cfg.items()
            if k not in self._SPECIAL_CFG_KEYS
        }
        # NOTE setup_figure and save_figure are handled separately

        # Go over all possible coordinates
        for ax_coords in self.coords_iter(match="all"):
            # Store a copy of the base configuration for these coordinates
            cfg[ax_coords] = copy.deepcopy(self._base_cfg)

            # Go over the list of updates and apply them
            for ax_key, update_params in self._axis_specific_updates.items():
                if any(k in self._FIGURE_HELPERS for k in update_params):
                    raise PlotConfigError(
                        "Cannot set axis-specific configuration for figure-"
                        "level helpers. Remove any keys that refer to figure-"
                        f"level helpers ({', '.join(self._FIGURE_HELPERS)})\n"
                        f"Given parameters were:\n  {update_params}."
                    )

                # Work on a copy
                update_params = copy.deepcopy(update_params)

                # Determine which axis or axes to update
                if "axis" in update_params:
                    axis = update_params.pop("axis")
                elif isinstance(ax_key, tuple) and len(ax_key) == 2:
                    axis = ax_key
                else:
                    raise PlotConfigError(
                        "No axis could be determined for axis-specific update "
                        f"'{ax_key}'! Need either a 2-tuple for the update "
                        "key itself or an explicit `axis` key that specifies "
                        "which axes to match the update to."
                    )

                # Check if there is a match
                if not _coords_match(
                    ax_coords, match=axis, full_shape=self.axes.shape
                ):
                    # Nothing to update for this coordinate
                    continue

                # In-place update the parameter dict. Copies have been created
                # above already, so no need to do this here again.
                recursive_update(cfg[ax_coords], update_params)

        # Now created configurations for all axes and applied all updates
        return cfg

    def _raise_on_invalid_helper_name(
        self, *helper_names: str, special_cfg_keys_allowed: bool = False
    ):
        """Makes sure the given helper names are valid.

        Args:
            *helper_names (str): Names of the helpers to check
            special_cfg_keys_allowed (bool, optional): Whether to regard the
                special configuration keys as valid or not.

        Raises:
            ValueError: On invalid helper name
        """
        valid_names = self.available_helpers

        if special_cfg_keys_allowed:
            valid_names += self._SPECIAL_CFG_KEYS

        for helper_name in helper_names:
            if helper_name in valid_names:
                # Is valid
                continue

            # Not valid!
            raise ValueError(
                "No helper with name '{}' available! "
                "Available helpers: {}. "
                "Special configuration keys ({}) {} allowed for "
                "this operation."
                "".format(
                    helper_name,
                    ", ".join(self.available_helpers),
                    ", ".join(self._SPECIAL_CFG_KEYS),
                    ("are also" if special_cfg_keys_allowed else "are NOT"),
                )
            )

    def _handle_errors(self, *errors, raise_on_error: bool = None):
        """Helper method to handle errors"""
        if not errors:
            return

        phe = PlotHelperErrors(*errors)
        if self._raise_on_error or raise_on_error:
            raise phe
        log.warning(phe)

    def _find_axis_coords(self, ax) -> Tuple[int, int]:
        """Find the coordinates of the given axis object in the axes array"""
        try:
            col, row = np.where(self.axes == ax)
            col, row = col[0], row[0]

        except Exception as exc:
            raise ValueError(
                "Could not find the given axis within the associated axes "
                "array! Is the correct figure associated?"
            ) from exc

        return col, row

    # -------------------------------------------------------------------------
    # -- Helper methods -------------------------------------------------------
    # -------------------------------------------------------------------------
    # .........................................................................
    # ... acting on the figure

    def _hlpr_align_labels(self, *, x: bool = True, y: bool = True):
        """Aligns axis labels in the whole figure by calling
        :py:meth:`matplotlib.figure.Figure.align_xlabels` and/or
        :py:meth:`matplotlib.figure.Figure.align_ylabels`.
        """
        if x:
            self.fig.align_xlabels()
        if y:
            self.fig.align_ylabels()

    def _hlpr_set_suptitle(
        self,
        *,
        title: str = None,
        margin: float = 0.025,
        **title_kwargs,
    ):
        """Set the figure title using
        :py:meth:`matplotlib.figure.Figure.suptitle`.

        This figure-level helper automatically vertically adjusts the subplot
        sizes to fit the suptitle into the figure without overlapping. This is
        *not* done if the ``title`` string is empty or if the y-position is
        specified via the ``y`` argument. When repetitively invoking this
        helper on the same figure, the subplot sizes are re-adjusted each time.

        Args:
            title (str, optional): The title to be set
            margin (float, optional): An additional vertical margin between
                the figure and the suptitle.
            **title_kwargs: Passed on to ``fig.suptitle``
        """
        st = self.fig.suptitle(title, **title_kwargs)

        if title is not None and not "y" in title_kwargs:
            # Make some figure adjustments such that it does not overlap with
            # the already existing parts.
            _, space_needed = calculate_space_needed_hv(self.fig, st)
            self.fig.subplots_adjust(top=(1.0 - (space_needed + margin)))

    def _hlpr_set_figlegend(
        self,
        *,
        gather_from_fig: bool = False,
        custom_labels: Sequence[str] = (),
        hiding_threshold: int = None,
        loc="center right",
        margin: float = 0.015,
        **legend_kwargs,
    ):
        """Sets a figure legend.

        As a source of handles and labels, uses all those tracked via
        :py:meth:`.track_handles_labels`.
        Furthermore, ``gather_from_fig`` controls whether to additionally
        retrieve already existing handles and labels from any legend used
        within the figure, e.g. on all of the axes.

        For legend locations on the **right** side of the figure, this will
        additionally adjust the subplot shape to accomodate for the figure
        legend without overlapping. This is not done for other ``loc`` values.

        If no handles could be retrieved by the above procedure, no figure
        legend will be added.

        Args:
            gather_from_fig (bool, optional): Whether to extract figure handles
                and labels from already existing legends used in the figure.
            custom_labels (Sequence[str], optional): Custom labels to assign
                for the given handles.
            hiding_threshold (int, optional): If there are more handles or
                labels available than the threshold, no figure legend will be
                added.
            loc (str, optional): The location of the figure legend. Note that
                figure adjustments are done for a figure legend on the right
                side of the figure; these are not done if the figure is located
                somewhere else, which might lead to it overlapping with other
                axes.
            margin (float, optional): An additional horizontal margin between
                the figure legend and the axes.
            **legend_kwargs: Passed on to
                :py:meth:`matplotlib.figure.Figure.legend`.

        Raises:
            RuntimeError: If a figure legend was already set via the helper.
        """
        # Ensure that this is not called multiple times, because it's quite
        # costly to do all of the below (and can really mess up the figure
        # if applied multiple times).
        if self._figlegend:
            raise RuntimeError(
                "A figure legend was already set for this figure! "
                "Will not set another one."
            )

        # Now get the handles and labels . . . . . . . . . . . . . . . . . . .
        h, l = self.all_handles_labels

        if gather_from_fig:
            _h, _l = gather_handles_labels(self.fig)
            log.debug(
                "Gathered %d handles and labels from the whole figure.",
                len(_h),
            )
            h += _h
            l += _l

        h, l = remove_duplicate_handles_labels(h, l)

        # Evaluate custom labels and hiding threshold
        h, l, past_thresh = prepare_legend_args(
            h,
            l,
            custom_labels=custom_labels,
            hiding_threshold=hiding_threshold,
        )
        if not h or past_thresh:
            return

        # Construct the legend and keep track of it
        figlegend = self.fig.legend(h, l, loc=loc, **legend_kwargs)
        self._figlegend = figlegend

        # Figure adjustments . . . . . . . . . . . . . . . . . . . . . . . . .
        # TODO Extend to all locs
        if "right" in loc:
            # These are necessary to have enough space for the legend.
            space_needed, _ = calculate_space_needed_hv(self.fig, figlegend)
            self.fig.subplots_adjust(right=(1.0 - (space_needed + margin)))

    def _hlpr_subplots_adjust(self, **kwargs):
        """Invokes :py:meth:`matplotlib.figure.Figure.subplots_adjust` on the
        whole figure.
        """
        self.fig.subplots_adjust(**kwargs)

    def _hlpr_figcall(self, *, functions: Sequence[dict], **shared_kwargs):
        """Figure-level helper that can be used to call multiple functions.
        This helper is invoked *before* the axis-level helper.

        See :py:meth:`._hlpr_call` for more information and examples.

        Args:
            functions (Sequence[dict]): A sequence of function call
                specifications. Each dict needs to contain at least the key
                ``function`` which determines which function to invoke. Further
                arguments are parsed into the positional and keyword arguments
                of the to-be-invoked function.
            **shared_kwargs: Passed on as keyword arguments to *all* function
                calls in ``functions``.
        """
        self._hlpr_call(functions=functions, **shared_kwargs)

    def _hlpr_autofmt_xdate(self, **kwargs):
        """Invokes :py:meth:`~matplotlib.figure.Figure.autofmt_xdate` on the
        figure, passing all arguments along. This can be useful when x-labels
        are overlapping (e.g. because they are dates).
        """
        self.fig.autofmt_xdate(**kwargs)

    # .........................................................................
    # ... acting on a single axis

    def _hlpr_set_title(self, *, title: str = None, **title_kwargs):
        """Sets the title of the current axes.

        Args:
            title (str, optional): The title to be set
            **title_kwargs: Passed on to
                :py:meth:`matplotlib.axes.Axes.set_title`
        """
        self.ax.set_title(title, **title_kwargs)

    def _hlpr_set_labels(
        self,
        *,
        x: Union[str, dict] = None,
        y: Union[str, dict] = None,
        z: Union[str, dict] = None,
        only_label_outer: bool = False,
        rotate_z_label: bool = None,
    ):
        """Sets the x, y, and z label of the current axis.

        Args:
            x (Union[str, dict], optional): Either the label as a string or
                a dict with key ``label``, where all further keys are passed on
                to :py:meth:`matplotlib.axes.Axes.set_xlabel`
            y (Union[str, dict], optional): Either the label as a string or
                a dict with key ``label``, where all further keys are passed on
                to :py:meth:`matplotlib.axes.Axes.set_ylabel`
            z (Union[str, dict], optional): Either the label as a string or
                a dict with key ``label``, where all further keys are passed on
                to :py:meth:`mpl_toolkits.mplot3d.axes3d.Axes3D.set_zlabel`.
                If there is no z-axis, this will be silently ignored.
            only_label_outer (bool, optional): If True, call
                :py:meth:`matplotlib.axes.Axes.label_outer` such that
                only tick labels on "outer" axes are visible:
                x-labels are only kept for subplots on the last row; y-labels
                only for subplots on the first column. Note that this applies
                to both axes and may lead to existing axes being hidden.
            rotate_z_label (bool, optional): If given, sets
                :py:meth:`mpl_toolkits.mplot3d.axis3d.Axis.set_rotate_label`.
                If there is no z-axis, this will be silently ignored.
        """

        def set_label(func: Callable, *, label: str = None, **label_kwargs):
            return func(label, **label_kwargs)

        if x is not None:
            x = x if not isinstance(x, str) else dict(label=x)
            set_label(self.ax.set_xlabel, **x)

        if y is not None:
            y = y if not isinstance(y, str) else dict(label=y)
            set_label(self.ax.set_ylabel, **y)

        if hasattr(self.ax, "zaxis"):
            if z is not None:
                z = z if not isinstance(z, str) else dict(label=z)
                set_label(self.ax.set_zlabel, **z)

            if rotate_z_label is not None:
                self.ax.zaxis.set_rotate_label(rotate_z_label)

        if only_label_outer:
            self.ax.label_outer()

    def _hlpr_set_limits(
        self,
        *,
        x: Union[Sequence[Union[float, str]], dict] = None,
        y: Union[Sequence[Union[float, str]], dict] = None,
        z: Union[Sequence[Union[float, str]], dict] = None,
    ):
        """Sets the x, y, and z limits for the current axis. Allows some
        convenience definitions for the arguments and then calls
        :py:meth:`matplotlib.axes.Axes.set_xlim` and/or
        :py:meth:`matplotlib.axes.Axes.set_ylim`.
        :py:meth:`mpl_toolkits.mplot3d.axes3d.Axes3D.set_zlim`.

        The ``x``, ``y``, and ``z`` arguments can have the following form:

            - None:         Limits are not set
            - sequence:     Specify lower and upper values
            - dict:         Expecting keys ``lower`` and/or ``upper``

        The sequence or dict values can be:

            - None          Set automatically / do not set
            - numeric       Set to this value explicitly
            - ``min``       Set to the data minimum value on that axis
            - ``max``       Set to the data maximum value on that axis

        Args:
            x (Union[Sequence[Union[float, str]], dict], optional): The limits
                to set on the x-axis
            y (Union[Sequence[Union[float, str]], dict], optional): The limits
                to set on the y-axis
            z (Union[Sequence[Union[float, str]], dict], optional): The limits
                to set on the z-axis
                If there is no z-axis, this will be silently ignored.
        """

        def parse_args(
            args: Union[Sequence[Union[float, str]], dict], *, ax
        ) -> Tuple[float, float]:
            """Parses the limit arguments."""

            def parse_arg(arg: Union[float, str]) -> Union[float, None]:
                """Parses a single limit argument to either be float or None"""
                if not isinstance(arg, str):
                    # Nothing to parse
                    return arg

                if arg == "min":
                    arg = ax.get_data_interval()[0]
                elif arg == "max":
                    arg = ax.get_data_interval()[1]
                else:
                    raise PlotConfigError(
                        f"Got an invalid str-type argument '{arg}' to the "
                        "set_limits helper. Allowed: 'min', 'max', None, or a "
                        "numerical value specifying the lower or upper limit."
                    )

                # Check that it is finite
                if not np.isfinite(arg):
                    raise PlotConfigError(
                        "Could not get a finite value from the axis data to "
                        "use for setting axis limits to 'min' or 'max', "
                        "presumably because the axis is still empty."
                    )

                return arg

            # Special case: dict
            if isinstance(args, dict):
                # Make sure there are only allowed keys
                if [k for k in args.keys() if k not in ("lower", "upper")]:
                    raise PlotConfigError(
                        "There are invalid keys present in a dict-type "
                        "argument to set_limits! Only accepting keys 'lower' "
                        f"and 'upper', but got: {args}"
                    )

                lower = args.get("lower", None)
                upper = args.get("upper", None)

            else:
                lower, upper = args  # ... assuming sequence of length 2

            # Parse individually, then return
            return (parse_arg(lower), parse_arg(upper))

        # Now set the limits, using the helper functions defined above
        if x is not None:
            self.ax.set_xlim(*parse_args(x, ax=self.ax.xaxis))

        if y is not None:
            self.ax.set_ylim(*parse_args(y, ax=self.ax.yaxis))

        if hasattr(self.ax, "zaxis"):
            if z is not None:
                self.ax.set_zlim(*parse_args(z, ax=self.ax.zaxis))

    def _hlpr_set_margins(
        self,
        *,
        margins: Union[float, Tuple[float, float]] = None,
        x: float = None,
        y: float = None,
        tight: bool = True,
    ):
        """Sets the axes' margins via :py:meth:`matplotlib.axes.Axes.margins`.

        The padding added to each limit of the Axes is the margin times the
        data interval.
        All input parameters must be floats within the range [0, 1].

        Specifying any margin changes only the autoscaling; for example, if
        ``xmargin`` is not None, then ``xmargin`` times the X data interval
        will be added to each end of that interval before it is used in
        autoscaling.

        Args:
            margins (Union[float, Tuple[float, float]], optional): If a scalar
                argument is provided, it specifies both margins of the x-axis
                and y-axis limits. If a list- or tuple-like positional
                argument is provided, they will be interpreted as ``xmargin``,
                and ``ymargin``. If setting the margin on a single axis is
                desired, use the keyword arguments described below.
            x, y (float, optional): Specific margin values for the x-axis and
                y-axis, respectively. These cannot be used in combination with
                the ``margins`` argument, but can be used individually to
                alter on e.g., only the y-axis.
            tight (bool, optional): The tight parameter is passed to
                :py:meth:`matplotlib.axes.Axes.autoscale_view`, which is
                executed after a margin is changed; the default here is True,
                on the assumption that when margins are specified, no
                additional padding to match tick marks is usually desired. Set
                tight to None will preserve the previous setting.
        """
        kwargs = dict(x=x, y=y, tight=tight)

        if margins is not None:
            if x is not None or y is not None:
                raise TypeError(
                    "Cannot pass both `margins` and `x`/`y` arguments!"
                )

            if isinstance(margins, (float, int)):
                self.ax.margins(margins, **kwargs)
            else:
                self.ax.margins(*margins, **kwargs)
        else:
            self.ax.margins(**kwargs)

    def _hlpr_set_legend(
        self,
        *,
        use_legend: bool = True,
        gather_from_fig: bool = False,
        custom_labels: Sequence[str] = (),
        hiding_threshold: int = None,
        **legend_kwargs,
    ):
        """Sets a legend for the current axis.

        As a first step, this helper tries to extract all relevant legend
        handles and labels. If a legend was set previously and *no* handles and
        labels could be extracted in the typical way (i.e., using the
        ``ax.get_legend_handles_labels`` method) it will be attempted to
        retrieve them from existing :py:class:`matplotlib.legend.Legend`
        objects on the current axis.
        If ``gather_from_fig`` is given, the *whole* figure will be inspected,
        regardless of whether handles were found previously.

        Additionally, all axis-specific handles and labels tracked via
        :py:meth:`.track_handles_labels` will always be added.

        .. note::

            - If no handles can be found, the legend is hidden, also meaning
              that the ``legend_kwargs`` will not be passed on.
            - During gathering of handles and labels from the current axis or
              the figure, duplicates will be removed; duplicates are detected
              via their label strings, *not* via their handle.

        .. hint::

            To set a figure-level legend, use the ``set_figlegend`` helper.

        Args:
            use_legend (bool, optional): Whether to set a legend or not. If
                False, the legend will be removed.
            gather_from_fig (bool, optional): If set, will gather legend
                handles and labels from the whole figure. This can be useful to
                set if the relevant information is found on another axis or in
                a figure legend.
            custom_labels (Sequence[str], optional): If given, use these labels
                and associate them with existing labels. Note that if fewer
                labels are given than handles are available, those without a
                label will not be drawn.
            hiding_threshold (int, optional): If given, will hide legends
                that have more than this number of handles registered.
            **legend_kwargs: Passed on to ``ax.legend``
        """

        def hide_legend():
            legend = self.ax.legend((), fancybox=False, frameon=False)
            legend.set_visible(False)

        # Warn about usage of the old interface.
        # TODO Remove in Version 1.0
        if legend_kwargs.pop("use_figlegend", None):
            log.warning(
                "The `use_figlegend` argument is no longer supported "
                "for the `set_legend` helper. Use `set_figlegend` instead."
            )

        # Do explicit hiding first, don't need to do all the rest in this case
        if not use_legend:
            log.remark("Hiding legend ...")
            hide_legend()
            return

        # Try to get (dangling) handles and labels from the axis
        h, l = self.ax.get_legend_handles_labels()
        # NOTE Will be empty if ax.legend() was called already

        # Add those that have been tracked explicitly
        _h, _l = self.axis_handles_labels
        h += _h
        l += _l

        # If there were no handles available in this way, try to gather the
        # information from the current axis or the whole figure
        if not h or gather_from_fig:
            _h, _l = gather_handles_labels(
                self.fig if gather_from_fig else self.ax
            )
            log.debug(
                "Gathered %d handles and labels from %s.",
                len(_h),
                (
                    "the whole figure"
                    if (not h and gather_from_fig)
                    else "current axis"
                ),
            )

            h += _h
            l += _l

        # Remove potential duplicate handles (identified by the labels)
        h, l = remove_duplicate_handles_labels(h, l)

        # ... and evaluate the custom labels and hiding threshold
        h, l, past_thresh = prepare_legend_args(
            h,
            l,
            custom_labels=custom_labels,
            hiding_threshold=hiding_threshold,
        )

        # Hide or draw the legend
        if past_thresh or not h:
            hide_legend()

        else:
            self.ax.legend(h, l, **legend_kwargs)

    def _hlpr_set_texts(self, *, texts: Sequence[dict]):
        """Sets multiple text elements for the current axis.

        Example configuration:

        .. code-block:: yaml

            set_texts:
              texts:
                - x: 0
                  y: 1
                  s: some text
                  # ... more arguments to plt.text

        Args:
            texts (Sequence[dict]): A sequence of text specifications, that are
                passed to :py:func:`matplotlib.pyplot.text`
        """
        for kwargs in texts:
            self.ax.text(**kwargs)

    def _hlpr_annotate(self, *, annotations: Sequence[dict]):
        """Sets multiple annotations for the current axis.

        Example configuration:

        .. code-block:: yaml

            annotate:
              annotations:
                - xy: [1, 3.14159]
                  text: this is π
                  xycoords: data
                  # ... more arguments to plt.annotate
                - xy: [0, 0]
                  xycoords: data
                  text: this is zero
                  xytext: [0.1, 0.1]
                  arrowprops:
                    facecolor: black
                    shrink: 0.05

        Args:
            annotations (Sequence[dict]): A sequence of annotation parameters
                which will be passed to :py:func:`matplotlib.pyplot.annotate`
        """
        for kwargs in annotations:
            self.ax.annotate(**kwargs)

    def _hlpr_set_hv_lines(self, *, hlines: list = None, vlines: list = None):
        """Sets one or multiple horizontal or vertical lines using
        :py:meth:`matplotlib.axes.Axes.axhline` and / or
        :py:meth:`matplotlib.axes.Axes.axvline`.

        Args:
            hlines (list, optional): list of numeric positions of the lines or
                or list of dicts with key ``pos`` determining the position, key
                ``limits`` determining the relative limits of the line, and all
                additional arguments being passed on to the matplotlib
                function.
            vlines (list, optional): list of numeric positions of the lines or
                or list of dicts with key ``pos`` determining the position, key
                ``limits`` determining the relative limits of the line, and all
                additional arguments being passed on to the matplotlib
                function.
        """

        def set_line(
            func: Callable,
            *,
            pos: float,
            limits: tuple = (0.0, 1.0),
            **line_kwargs,
        ):
            try:
                pos = float(pos)

            except Exception as err:
                raise PlotConfigError(
                    f"Got non-numeric value '{pos}' for `pos` argument in "
                    "set_hv_lines helper!"
                )

            func(pos, *limits, **line_kwargs)

        if hlines is not None:
            for line_spec in hlines:
                if isinstance(line_spec, dict):
                    set_line(self.ax.axhline, **line_spec)
                else:
                    set_line(self.ax.axhline, pos=line_spec)

        if vlines is not None:
            for line_spec in vlines:
                if isinstance(line_spec, dict):
                    set_line(self.ax.axvline, **line_spec)
                else:
                    set_line(self.ax.axvline, pos=line_spec)

    def _hlpr_set_scales(
        self,
        *,
        x: Union[str, dict] = None,
        y: Union[str, dict] = None,
        z: Union[str, dict] = None,
    ):
        """Sets the scales for the current axis

        The arguments are used to call
        :py:meth:`matplotlib.axes.Axes.set_xscale` and/or
        :py:meth:`matplotlib.axes.Axes.set_yscale` and
        :py:meth:`mpl_toolkits.mplot3d.axes3d.Axes3D.set_zscale`, respectively.
        For string-like arguments, the value is directly used to set the scale
        for that axis, e.g. ``linear``, ``log``, ``symlog``.
        Otherwise, dict-like arguments are expected where a ``scale`` key is
        present and defines which type of scale to use. All further arguments
        are passed on; these are relevant for the symmetrical logarithmic
        scale, for example.

        Args:
            x (Union[str, dict], optional): The scales to use on the x-axis
            y (Union[str, dict], optional): The scales to use on the y-axis
            z (Union[str, dict], optional): The scales to use on the z-axis.
                If there is no z-axis, this will be silently ignored.
        """

        def set_scale(func: Callable, *, scale: str = None, **scale_kwargs):
            func(scale, **scale_kwargs)

        if x:
            x = x if not isinstance(x, str) else dict(scale=x)
            set_scale(self.ax.set_xscale, **x)

        if y:
            y = y if not isinstance(y, str) else dict(scale=y)
            set_scale(self.ax.set_yscale, **y)

        if hasattr(self.ax, "zaxis"):
            if z:
                z = z if not isinstance(z, str) else dict(scale=z)
                set_scale(self.ax.set_zscale, **z)

    def _hlpr_set_ticks(
        self,
        *,
        x: Union[list, dict] = None,
        y: Union[list, dict] = None,
        z: Union[list, dict] = None,
    ):
        """Sets the ticks for the current axis

        The arguments are used to call
        :py:meth:`matplotlib.axes.Axes.set_xticks` or
        :py:meth:`matplotlib.axes.Axes.set_yticks` or
        :py:meth:`mpl_toolkits.mplot3d.axes3d.Axes3D.set_zticks`,
        and :py:meth:`matplotlib.axes.Axes.set_xticklabels` or
        :py:meth:`matplotlib.axes.Axes.set_yticklabels` or
        :py:meth:`mpl_toolkits.mplot3d.axes3d.Axes3D.set_zticklabels`,
        respectively.

        The dict-like arguments may contain the keys ``major`` and/or
        ``minor``, referring to major or minor tick locations and labels,
        respectively.
        They should either be list-like, directly specifying the ticks'
        locations, or dict-like requiring a ``locs`` key that contains the
        ticks' locations and is passed on to the ``set_<x/y/z>ticks`` call.
        as ``ticks`` argument. Further kwargs such as ``labels`` can be given
        and are passed on to the ``set_<x/y/z>ticklabels`` call.

        Example:

        .. code-block:: yaml

            set_ticks:
              x:
                major: [2, 4, 6, 8]
                minor:
                  locs: [1, 3, 5, 7, 9]
              y:
                major:
                  locs: [0, 1000, 2000, 3000]
                  labels: [0, 1k, 2k, 3k]
                  # ... further kwargs here specify label aesthetics

              z: [-1, 0, +1]  # same as z: {major: {locs: [-1, 0, +1]}}

        Args:
            x (Union[list, dict], optional): The ticks and optionally their
                labels to set on the x-axis
            y (Union[list, dict], optional): The ticks and optionally their
                labels to set on the y-axis
            z (Union[list, dict], optional): The ticks and optionally their
                labels to set on the z-axis.
                If there is no z-axis, this will be silently ignored.
        """

        def set_ticks(
            func: Callable,
            *,
            ticks: list,
            minor: bool,
        ):
            func(ticks, minor=minor)

        def set_ticklabels(
            func: Callable, *, labels: list, **ticklabels_kwargs
        ):
            func(labels, **ticklabels_kwargs)

        def set_ticks_and_labels(
            *, axis: str, minor: bool, axis_cfg: Union[dict, list, tuple]
        ):
            axis_cfg = (
                axis_cfg
                if not isinstance(axis_cfg, (list, tuple))
                else dict(locs=axis_cfg)
            )

            if axis_cfg.get("labels") and not axis_cfg.get("locs"):
                raise PlotConfigError(
                    "Labels can only be set through the `labels` argument "
                    "if their location is also given through the `locs` "
                    "argument. However, `locs` is not given."
                )

            # Get the axis specific
            if axis == "x":
                ticks_func = self.ax.set_xticks
                tickslabel_func = self.ax.set_xticklabels
            elif axis == "y":
                ticks_func = self.ax.set_yticks
                tickslabel_func = self.ax.set_yticklabels
            elif axis == "z":
                ticks_func = self.ax.set_zticks
                tickslabel_func = self.ax.set_zticklabels

            # Set the ticks
            set_ticks(ticks_func, ticks=axis_cfg.pop("locs"), minor=minor)

            # Set the tick labels, passing on the additional kwargs
            if axis_cfg.get("labels"):
                set_ticklabels(tickslabel_func, minor=minor, **axis_cfg)

        if x:
            if isinstance(x, (list, tuple)):
                x = dict(major=dict(locs=x))

            if x.get("major") is not None:
                set_ticks_and_labels(
                    axis="x", minor=False, axis_cfg=x["major"]
                )

            if x.get("minor") is not None:
                set_ticks_and_labels(axis="x", minor=True, axis_cfg=x["minor"])

        if y:
            if isinstance(y, (list, tuple)):
                y = dict(major=dict(locs=y))

            if y.get("major") is not None:
                set_ticks_and_labels(
                    axis="y", minor=False, axis_cfg=y["major"]
                )

            if y.get("minor") is not None:
                set_ticks_and_labels(axis="y", minor=True, axis_cfg=y["minor"])

        if hasattr(self.ax, "zaxis"):
            if z:
                if isinstance(z, (list, tuple)):
                    z = dict(major=dict(locs=z))

                if z.get("major") is not None:
                    set_ticks_and_labels(
                        axis="z", minor=False, axis_cfg=z["major"]
                    )

                if z.get("minor") is not None:
                    set_ticks_and_labels(
                        axis="z", minor=True, axis_cfg=z["minor"]
                    )

    def _hlpr_set_tick_locators(
        self, *, x: dict = None, y: dict = None, z: dict = None
    ):
        """Sets the tick locators for the current axis

        The arguments are used to call
        ``ax.{x,y, z}axis.set_{major/minor}_locator``, respectively.
        The dict-like arguments must contain the keys ``major`` and/or
        ``minor``, referring to major or minor tick locators. These need to
        specify a name that is looked up in :py:mod:`matplotlib.ticker`.
        They can contain a list-like ``args`` keyword argument that defines
        the arguments to pass on as positional args to the called function.
        Further kwargs are passed on to
        ``ax.{x,y, z}axis.set_{major/minor}_locator``.

        Example:

        .. code-block:: yaml

            set_tick_locators:
              x:
                major:
                  name: MaxNLocator    # looked up from matplotlib.ticker
                  nbins: 6
                  integer: true
                  min_n_ticks: 3
              y:
                major:
                  name: MultipleLocator
                  args: [2]


        For more information, see:

            - https://matplotlib.org/gallery/ticks_and_spines/tick-locators.html
            - https://matplotlib.org/api/_as_gen/matplotlib.axis.Axis.set_major_locator.html
            - https://matplotlib.org/api/_as_gen/matplotlib.axis.Axis.set_minor_locator.html

        Args:
            x (dict, optional): The configuration of the x-axis tick locator
            y (dict, optional): The configuration of the y-axis tick locator
            z (dict, optional): The configuration of the z-axis tick locator.
                If there is no z-axis, this will be silently ignored.
        """
        set_tick_locators_or_formatters(
            ax=self.ax, kind="locator", x=x, y=y, z=z
        )

    def _hlpr_set_tick_formatters(
        self, *, x: dict = None, y: dict = None, z: dict = None
    ):
        """Sets the tick formatters for the current axis

        The arguments are used to call
        ``ax.{x,y, z}axis.set_{major/minor}_formatter``, respectively. The
        dict-like arguments must contain the keys ``major`` and/or ``minor``,
        referring to major or minor tick formatters. These need to specify a
        name that is looked up in :py:mod:`matplotlib.ticker`.
        They can contain a list-like ``args`` keyword argument that defines
        the arguments to pass on as positional args to the called function.
        Further kwargs are passed on to
        ``ax.{x,y, z}axis.set_{major/minor}_formatter``.

        Example:

        .. code-block:: yaml

            set_tick_formatters:
              x:
                major:
                  name: StrMethodFormatter
                  args: ['{x:.3g}']
              y:
                major:
                  name: FuncFormatter
                  args: [!dag_result my_formatter_lambda]
                  # any kwargs here passed also to FuncFormatter

              z:
                major:
                  name: DateFormatter  # looked up from matplotlib.dates
                  args: ["%H:%M:%S"]  # or "%Y-%m-%d"

        For more information, see:

            - https://matplotlib.org/gallery/ticks_and_spines/tick-formatters.html
            - https://matplotlib.org/api/_as_gen/matplotlib.axis.Axis.set_major_formatter.html
            - https://matplotlib.org/api/_as_gen/matplotlib.axis.Axis.set_minor_formatter.html

        Args:
            x (dict, optional): The configuration of the x-axis tick formatter
            y (dict, optional): The configuration of the y-axis tick formatter
            z (dict, optional): The configuration of the z-axis tick formatter.
                If there is no z-axis, this will be silently ignored.
        """
        set_tick_locators_or_formatters(
            ax=self.ax, kind="formatter", x=x, y=y, z=z
        )

    def _hlpr_call(
        self,
        *,
        functions: Sequence[dict],
        funcs_lookup_dict: Dict[str, Callable] = None,
        **shared_kwargs,
    ):
        """Axis-level helper that can be used to call multiple functions.

        Functions can be specified in three ways:

            - as string, being looked up from a pre-defined dict
            - as 2-tuple ``(module, name)`` which will be imported on the fly
            - as callable, which will be used directly

        The implementation of this is shared with the plot function
        :py:func:`~dantro.plot.funcs.multiplot.multiplot`. See
        there for more information.

        The figure-level helper ``figcall`` is identical to this helper, but is
        invoked *before* the axis-specific helpers are invoked.

        .. hint::

            To pass custom callables, use the data transformation framework and
            the ``!dag_result`` placeholder, see :ref:`dag_result_placeholder`.

        .. note::

            While most matplotlib-based functions will automatically operate on
            the current axis, some function calls may require an axis object.
            If so, use the ``pass_axis_object_as`` argument, which specifies
            the name of the keyword argument as which the current axis is
            passed to the function call.

        Example:

        .. code-block:: yaml

            call:
              functions:
                # Look up function from dict, containing common seaborn and
                # pyplot plotting functions (see multiplot for more info)
                - function: sns.lineplot
                  data: !dag_result my_custom_data

                # Import function via `(module, name)` specification
                - function: [matplotlib, pyplot.subplots_adjust]
                  left: 0.1
                  right: 0.9

                # Pass a custom callable, selected via DAG framework
                - function: !dag_result my_callable
                  args: [foo, bar]
                  # ... keyword arguments here

                # Pass current axis object as keyword argument
                - function: !dag_result my_callable_operating_on_ax
                  pass_axis_object_as: ax

                # Pass helper object itself as keyword argument
                - function: !dag_result my_callable_operating_on_helper
                  pass_helper: true

        Args:
            functions (Sequence[dict]): A sequence of function call
                specifications. Each dict needs to contain at least the key
                ``function`` which determines which function to invoke. Further
                arguments are parsed into the positional and keyword arguments
                of the to-be-invoked function.
            funcs_lookup_dict (Dict[str, Callable], optional): If given, will
                look up the function names from this dict instead of the
                default dict.
            **shared_kwargs: Passed on as keyword arguments to *all* function
                calls in ``functions``.
        """
        from .funcs._multiplot import parse_and_invoke_function

        for call_num, func_kwargs in enumerate(functions):
            parse_and_invoke_function(
                hlpr=self,
                funcs=funcs_lookup_dict,
                shared_kwargs=shared_kwargs,
                func_kwargs=func_kwargs,
                show_hints=False,
                call_num=call_num,
            )

    # .........................................................................
    # ... using seaborn

    def _hlpr_despine(self, **kwargs):
        """Despines the current *axis* using :py:func:`seaborn.despine`.

        To despine the whole *figure*, apply this helper to all axes.
        Refer to the seaborn documentation for available arguments:
        https://seaborn.pydata.org/generated/seaborn.despine.html

        Args:
            **kwargs: Passed on to ``seaborn.despine``.
        """
        if "fig" in kwargs:
            raise ValueError(
                "Got unexpected `fig` argument! To apply the `despine` helper "
                "to all axes, invoke the helper on each axis separately."
            )
        sns.despine(ax=self.ax, **kwargs)



"""Implements the :py:class:`~dantro.plot_mngr.PlotManager`, which handles the
configuration of multiple plots and prepares the data and configuration to pass
to the respective :ref:`plot creators <plot_creators>`.
See :ref:`the user manual <plot_manager>` for more information.
"""

import copy
import fnmatch
import logging
import os
import textwrap
import time
from collections import OrderedDict
from functools import partial
from typing import Any, Callable, Dict, List, Literal, Sequence, Tuple, Union

from paramspace import ParamDim, ParamSpace

from ._import_tools import get_resource_path
from .abc import BAD_NAME_CHARS as _BAD_NAME_CHARS
from .data_mngr import DataManager
from .exceptions import *
from .plot import BasePlotCreator, SkipPlot
from .plot._cfg import resolve_based_on as _resolve_based_on
from .plot._cfg import resolve_based_on_single as _resolve_based_on_single
from .plot._cfg import resolve_plot_cfgs_shortcuts as _resolve_shortcuts
from .plot.creators import ALL_PLOT_CREATORS as ALL_PCRS
from .plot.utils import PlotFuncResolver as _PlotFuncResolver
from .tools import format_time as _format_time
from .tools import load_yml, make_columns, recursive_update, write_yml

# .............................................................................

log = logging.getLogger(__name__)

_fmt_time = partial(_format_time, ms_precision=1)

BAD_PLOT_NAME_CHARS = tuple(
    [c for c in _BAD_NAME_CHARS if c not in ("/",)] + ["."]
)
"""Substrings that may not appear in plot names.

Unlike the :py:data:`~dantro.abc.BAD_NAME_CHARS`, these *allow* the ``/`` char
(such that new directories can be created) and disallows the ``.`` character
(in order to not get confused with file extensions).
"""

BASE_PLOTS_CFG_PATH: str = get_resource_path("dantro", "cfg/base_plots.yml")
"""The path to the base plot configurations pool for dantro.

If the ``use_dantro_base_cfg_pool`` flag is set when initializing a
:py:class:`~dantro.plot_mngr.PlotManager`, this file will be used as the first
entry in the sequence of config pools.

Also see :ref:`dantro_base_plots` for more information."""


# -----------------------------------------------------------------------------


class PlotManager:
    """The PlotManager takes care of configuring plots and calling the
    selected :ref:`plot creators <plot_creators>` that then actually carry out
    the plotting operation.

    It is a high-level class that is aware of a larger plot configuration and
    aggregates all general capabilities needed to configure and carry out plots
    using the plotting framework.

    See :ref:`the user manual <plot_manager>` for more information.
    """

    PLOT_FUNC_RESOLVER: type = _PlotFuncResolver
    """The class to use for resolving plot function objects"""

    CREATORS: Dict[str, type] = ALL_PCRS
    """The mapping of creator names to classes.
    By default, all available dantro plot creators are registered.

    When subclassing PlotManager and desiring to *extend* the creator mapping,
    use ``dict(**dantro.plot.creators.ALL, my_new_creator=MyNewCreator)``
    to include the default creator mapping."""

    DEFAULT_OUT_FSTRS: Dict[str, str] = dict(
        timestamp="%y%m%d-%H%M%S",
        #
        # representing the parameter space sweep state and its coordinate
        state_no="{no:0{digits:d}d}",
        state="{name:}_{val:}",
        state_name_replace_chars=[],  # (".", "-")
        state_val_replace_chars=[("/", "-")],
        state_join_char="__",
        state_vector_join_char="-",
        #
        # final fstr for single plot and config path
        path="{name:}{ext:}",
        plot_cfg="{basename:}_cfg.yml",
        #
        # and for sweep
        sweep="{name:}/{state_no:}__{state:}{ext:}",
        plot_cfg_sweep="{name:}/sweep_cfg.yml",
    )
    """The default values for the output format strings, used when composing
    the file name of a plot."""

    SPECIAL_BASE_CFG_POOL_LABELS: Sequence[str] = (
        "plot",
        "plot_from_cfg",
        "plot_from_cfg_unused",
        "plot_pspace",
    )
    """Special keys that may not be used as labels for the base configuration
    pools."""

    # . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . . .

    def __init__(
        self,
        *,
        dm: DataManager,
        default_plots_cfg: Union[dict, str] = None,
        out_dir: Union[str, None] = "{timestamp:}/",
        base_cfg_pools: Sequence[Tuple[str, Union[dict, str]]] = (),
        use_dantro_base_cfg_pool: bool = True,
        out_fstrs: dict = None,
        plot_func_resolver_init_kwargs: dict = None,
        shared_creator_init_kwargs: dict = None,
        creator_init_kwargs: Dict[str, dict] = None,
        default_creator: str = None,
        save_plot_cfg: bool = True,
        raise_exc: bool = False,
        cfg_exists_action: str = "raise",
    ):
        """Initialize a PlotManager, which provides a uniform configuration
        interface for creating plots and passes tasks on to the respective
        plot creators.

        To avoid copy-paste of plot configurations, the PlotManager comes with
        versatile capabilities to define default plots and re-use other plots.

        - The ``default_plots_cfg`` specifies plot configurations that are
          to be carried out by default when calling the plotting method
          :py:meth:`.plot_from_cfg`.
        - When calling any of the plot methods :py:meth:`.plot_from_cfg` or
          :py:meth:`.plot`, there is the possibility to update the existing
          configuration dict with new entries.
        - At each stage, the ``based_on`` feature allows to make a plot
          configuration inherit entries from an existing configuration.
          These are looked up from the ``base_cfg_pools`` following the
          rules described in :py:func:`~dantro.plot._cfg.resolve_based_on`.

        For more information on how the plot configuration can be defined, see
        :ref:`plot_cfg_inheritance`.

        Args:
            dm (DataManager): The DataManager-derived object to read the plot
                data from.
            default_plots_cfg (Union[dict, str], optional): The default plots
                config or a path to a YAML file to import. Used as defaults
                when calling :py:meth:`.plot_from_cfg`
            out_dir (Union[str, None], optional): If given, will use this
                output directory as basis for the output path for each plot.
                The path can be a format-string; it is evaluated upon call to
                the plot command. Available keys: ``timestamp``, ``name``, ...
                For a relative path, this will be relative to the DataManager's
                output directory. Absolute paths remain absolute.
                If this argument evaluates to False, the DataManager's output
                directory will be the output directory.
            base_cfg_pools (Sequence[Tuple[str, Union[dict, str]]], optional):
                The base configuration pools are used to perform the lookups of
                ``based_on`` entries, see :ref:`plot_cfg_inheritance`.
                The tuples in these sequence consist of ``(label, plots_cfg)``
                pairs and are fed to :py:meth:`.add_base_cfg_pool`; see there
                for more information.
            use_dantro_base_cfg_pool (bool, optional): If set, will use
                dantro's own base plot configuration pool as the *first* entry
                in the pool sequence. Refer to the
                :ref:`corresponding documentation page <dantro_base_plots>`
                for more information on available entries.
            out_fstrs (dict, optional): Format strings that define how the
                output path is generated. The dict given here updates the
                :py:attr:`.DEFAULT_OUT_FSTRS` class variable which holds the
                default values.

                Keys: ``timestamp`` (%-style), ``path``, ``sweep``, ``state``,
                ``plot_cfg``, ``state``, ``state_no``, ``state_join_char``,
                ``state_vector_join_char``.

                Available keys for ``path``: ``name``, ``timestamp``, ``ext``.

                Additionally, for ``sweep``: ``state_no``, ``state_vector``,
                    ``state``.

            plot_func_resolver_init_kwargs (dict, optional): Initialization
                arguments for the plot function resolver, by default
                :py:class:`~dantro.plot.utils.plot_func.PlotFuncResolver`.
            shared_creator_init_kwargs (dict, optional): Initialization
                arguments to the plot creator that are passed to *all* creators
                regardless of type (in contrast to ``creator_init_kwargs``).
            creator_init_kwargs (Dict[str, dict], optional): If given, these
                kwargs are passed to the initialization calls of the respective
                creator classes. These are resolved by the *names* given in the
                :py:attr:`.CREATORS` class variable and are passed to the
                :py:class:`~dantro.plot.creators.base.BasePlotCreator` or the
                respective derived class.
            default_creator (str, optional): If given, a plot without explicit
                ``creator`` declaration will use this creator as default.
            save_plot_cfg (bool, optional): If True, the plot configuration is
                saved to a yaml file alongside the created plot.
            raise_exc (bool, optional): Whether to raise exceptions if there
                are errors raised from the plot creator or errors in the plot
                configuration. If False, the errors will only be logged.
            cfg_exists_action (str, optional): Behaviour when a config file
                already exists. Can be: ``raise`` (default), ``skip``,
                ``append``, ``overwrite``, or ``overwrite_nowarn``.
        """
        # Public
        self.save_plot_cfg = save_plot_cfg
        self.raise_exc = raise_exc

        # Private or property-managed
        self._dm = dm
        self._out_dir = out_dir
        self._plot_info = []
        self._pfr_kwargs = (
            plot_func_resolver_init_kwargs
            if plot_func_resolver_init_kwargs
            else {}
        )
        self._shared_cckwargs = (
            shared_creator_init_kwargs if shared_creator_init_kwargs else {}
        )
        self._cckwargs = creator_init_kwargs if creator_init_kwargs else {}
        self._cfg_exists_action = cfg_exists_action
        self.default_creator = default_creator
        self._default_plots_cfg = None

        # Base configuration pools
        self._base_cfg_pools = OrderedDict()
        if use_dantro_base_cfg_pool:
            self.add_base_cfg_pool(
                label="dantro_base", plots_cfg=BASE_PLOTS_CFG_PATH
            )

        for _label, _plots_cfg in base_cfg_pools:
            self.add_base_cfg_pool(label=_label, plots_cfg=_plots_cfg)

        # Default plot configuration
        if default_plots_cfg:
            self._default_plots_cfg = self._prepare_cfg(default_plots_cfg)

        # Default format strings
        self._out_fstrs = self.DEFAULT_OUT_FSTRS
        if out_fstrs:
            self._out_fstrs = recursive_update(
                copy.deepcopy(self._out_fstrs), out_fstrs
            )

        log.debug("%s initialized.", self.__class__.__name__)
        log.debug(
            "Have %d base configuration pools available:  %s",
            len(self._base_cfg_pools),
            ", ".join(self._base_cfg_pools),
        )

    # .. Properties ...........................................................

    @property
    def out_fstrs(self) -> dict:
        """The dict of output format strings"""
        return self._out_fstrs

    @property
    def plot_info(self) -> List[dict]:
        """A list of dicts with info on all plots carried out so far"""
        return copy.deepcopy(self._plot_info)

    @property
    def base_cfg_pools(self) -> OrderedDict:
        """The base plot configuration pools, used for lookup the ``based_on``
        entry in plot configurations.

        The order of the entries in the pool is relevant, with later entries
        taking precedence over previous ones. See :ref:`plot_cfg_inheritance`
        for a more detailed description.
        """
        return self._base_cfg_pools

    @property
    def default_creator(self) -> str:
        """The name of the default creator"""
        return self._default_creator

    @default_creator.setter
    def default_creator(self, new_creator: str):
        """Set the name of the default creator, raising if it is invalid"""
        if new_creator and new_creator not in self.CREATORS:
            _avail = ", ".join(self.CREATORS.keys())
            raise InvalidCreator(
                f"No such creator '{new_creator}' available, only: {_avail}"
            )
        self._default_creator = new_creator

    # .. Configuration ........................................................

    def add_base_cfg_pool(self, *, label: str, plots_cfg: Union[str, dict]):
        """Adds a base configuration pool entry, allowing for the ``plots_cfg``
        to be a path to a YAML configuration file which is then loaded.

        The new pool is used for ``based_on`` lookups and takes precedence over
        existing entries. For more information on lookup rules, see
        :py:func:`~dantro.plot._cfg.resolve_based_on` and
        :ref:`plot_cfg_inheritance`.

        Args:
            label (str): A label of the pool that is used for identifying it.
            plots_cfg (Union[str, dict]): Description

        Raises:
            ValueError: If ``label`` already exists or is a special label.

        """
        if label in self._base_cfg_pools:
            raise ValueError(
                f"A base configuration labelled '{label}' already exists! "
                "Check if it was already added or choose a different name."
            )

        elif label in self.SPECIAL_BASE_CFG_POOL_LABELS:
            _special = ", ".join(self.SPECIAL_BASE_CFG_POOL_LABELS)
            raise ValueError(
                f"Invalid base configuration pool label '{label}'! Choosing "
                f"one of the special labels ({_special}) is not permitted."
            )

        self._base_cfg_pools[label] = self._prepare_cfg(plots_cfg)
        log.debug("Added base configuration pool '%s'.", label)

    # .. Helpers ..............................................................

    @staticmethod
    def _prepare_cfg(s: Union[str, dict]) -> Dict[str, dict]:
        """Prepares a plots configuration by either loading it from a YAML file
        if the given argument is a string or returning a deep copy of the given
        dict-like object.
        """
        if isinstance(s, str):
            plots_cfg = load_yml(s)
        else:
            plots_cfg = copy.deepcopy(s)

        if not plots_cfg:
            plots_cfg = {}
        elif not isinstance(plots_cfg, dict):
            raise TypeError(
                f"Got an unexpected plots configuration type:  {plots_cfg}"
            )

        # Resolve shortcut syntax
        plots_cfg = {
            k: _resolve_shortcuts(cfg, key=k) for k, cfg in plots_cfg.items()
        }

        return plots_cfg

    def _handle_exception(
        self,
        exc: Exception,
        *,
        pc: BasePlotCreator,
        debug: bool = None,
        ExcCls: type = PlottingError,
    ):
        """Helper for handling exceptions from the plot creator"""
        should_raise = debug or (debug is None and self.raise_exc)

        e_dbg = (
            "For a full error traceback, specify `debug: True` in the "
            "plot configuration or run the PlotManager in debug mode."
        )
        e_no_dbg = (
            "To ignore the error message and continue plotting with the "
            "other plots, specify `debug: False` in the plot "
            "configuration or disable debug mode for the PlotManager."
        )
        e_msg = (
            f"An error occurred during plotting with {pc.logstr}! "
            f"{e_dbg if not should_raise else e_no_dbg}\n\n"
            f"{exc.__class__.__name__}: {exc}"
        )

        if should_raise:
            raise ExcCls(e_msg) from exc
        log.error(e_msg)

    def _parse_out_dir(self, fstr: str, *, name: str) -> str:
        """Evaluates the format string to create an output directory path.

        Note that the directories are _not_ created; this is outsourced to the
        plot creator such that it happens as late as possible.

        Args:
            fstr (str): The format string to evaluate and create a directory at
            name (str): Name of the plot
            timestamp (float, optional): Description

        Returns:
            str: The path of the created directory
        """
        # Get date format string and current time and create a string
        timefstr = self._out_fstrs.get("timestamp", "%y%m%d-%H%M%S")
        timestr = time.strftime(timefstr)

        out_dir = str(fstr).format(timestamp=timestr, name=name)

        # Make sure it is absolute
        out_dir = os.path.expanduser(out_dir)
        if not os.path.isabs(out_dir):
            # Regard it as relative to the data manager's output directory
            out_dir = os.path.join(self._dm.dirs["out"], out_dir)

        # Return the full path
        return out_dir

    def _parse_out_path(
        self,
        creator: BasePlotCreator,
        *,
        name: str,
        out_dir: str,
        file_ext: str = None,
        state_no: int = None,
        state_no_max: int = None,
        state_vector: Tuple[int] = None,
        dims: dict = None,
    ) -> str:
        """Given a creator and (optionally) parameter sweep information, a full
        and absolute output path is generated, including the file extension.

        Note that the directories are _not_ created; this is outsourced to the
        plot creator such that it happens as late as possible.

        Args:
            creator (BasePlotCreator): The creator instance, used to
                extract information on the file extension.
            name (str): The name of the plot
            out_dir (str): The absolute output directory, prepended to all
                generated paths
            file_ext (str, optional): The file extension to use
            state_no (int, optional): The state number, starting with 0
            state_no_max (int, optional): The maximum state number
            state_vector (Tuple[int], optional): The state vector with info
                on how far each state dimension has progressed in the sweep
            dims (dict, optional): The dict of parameter dimensions of the
                sweep that is carried out.

        Returns:
            str: The fully parsed output path for this plot
        """

        def parse_state_pair(
            name: str, dim: ParamDim, *, fstrs: dict
        ) -> Tuple[str]:
            """Helper method to create a state pair"""
            # Parse the name
            for search, replace in fstrs["state_name_replace_chars"]:
                name = name.replace(search, replace)

            # Parse the value
            val = str(dim.current_value)
            for search, replace in fstrs["state_val_replace_chars"]:
                val = val.replace(search, replace)

            return name, val

        # Get the fstrs
        fstrs = self.out_fstrs

        # Evaluate the keys available for both cases
        keys = dict(timestamp=time.strftime(fstrs["timestamp"]), name=name)

        # Parse file extension and ensure it starts with a dot
        ext = file_ext if file_ext else creator.get_ext()

        if ext and ext[0] != ".":
            ext = "." + ext
        elif ext is None:
            ext = ""

        keys["ext"] = ext

        # Change behaviour depending on whether state information was given
        if state_no is None or (state_no == state_no_max == 0):
            # Assume the other arguments are also None -> Not part of the sweep
            # Evaluate it
            out_path = fstrs["path"].format(**keys)

        else:
            # Is part of a sweep
            # Parse additional keys
            # state number
            digits = len(str(state_no_max))
            keys["state_no"] = fstrs["state_no"].format(
                no=state_no, digits=digits
            )

            # state values -- need to do some parsing here ...
            state_pairs = [
                parse_state_pair(name, dim, fstrs=fstrs)
                for name, dim in dims.items()
            ]

            sjc = fstrs["state_join_char"]
            keys["state"] = sjc.join(
                [fstrs["state"].format(name=k, val=v) for k, v in state_pairs]
            )

            # state vector
            svjc = fstrs["state_vector_join_char"]
            keys["state_vector"] = svjc.join([str(s) for s in state_vector])

            # Evaluate it
            out_path = fstrs["sweep"].format(**keys)

        # Prepend the output directory and return
        out_path = os.path.join(out_dir, out_path)

        return out_path

    def _check_plot_name(self, name: str) -> None:
        """Raises if a plot name contains bad characters"""
        if any([s in name for s in BAD_PLOT_NAME_CHARS]):
            _bad_name_chars = ", ".join(repr(s) for s in BAD_PLOT_NAME_CHARS)
            raise ValueError(
                f"The plot name '{name}' contains unsupported characters! "
                "Remove any of the following offending substrings from the "
                f"plot function name: {_bad_name_chars}"
            )

    def _get_plot_func(self, **resolver_kwargs) -> Callable:
        """Instantiates a plot function resolver,
        :py:class:`~dantro.plot.utils.plot_func.PlotFuncResolver`, and uses
        it to get the desired plot function callable.
        """
        pf_resolver = self._get_plot_func_resolver(**self._pfr_kwargs)
        return pf_resolver.resolve(**resolver_kwargs)

    def _get_plot_func_resolver(self, **init_kwargs) -> _PlotFuncResolver:
        """Instantiates the plot function resolver object with the given
        initialization arguments.

        This method is called from :py:meth:`._get_plot_func` and can be used
        for more conveniently controlling how the resolver is set up.
        By default, the ``init_kwargs`` will be equivalent to the
        ``plot_func_resolver_init_kwargs`` given to :py:meth:`.__init__`.
        """
        return self.PLOT_FUNC_RESOLVER(**init_kwargs)

    def _get_plot_creator(
        self,
        *,
        creator: Union[str, Callable],
        plot_func: Callable,
        name: str,
        init_kwargs: dict,
    ) -> BasePlotCreator:
        """Determines which plot creator to use by looking at the given
        arguments and the plotting function.

        Then, sets up the corresponding creator and returns it.

        This method is called from :py:meth:`._plot`.

        Args:
            creator (Union[str, Callable]): The name of the creator to be
                looked up in :py:attr:`.CREATORS`. Can also be None, in
                which case it is attempted to look it up from the ``plot_func``
                's ``creator`` attribute. If that was not possible either, the
                :py:attr:`.default_creator` is used. If a callable is given,
                will use that as a factory to construct the creator instance.
            name (str): The name that will be used for the plot creator,
                typically the plot name itself.
            init_kwargs (dict): Additional creator initialization parameters

        Returns:
            BasePlotCreator: The selected creator object, fully initialized.
        """
        # If no creator is given, try to retrieve it by some other means
        if creator is None:
            log.debug("No creator specified.")
            if hasattr(plot_func, "creator"):
                creator = plot_func.creator
                log.debug(
                    "  Using creator specified by plot function instead:  %s",
                    creator,
                )

            elif self.default_creator:
                creator = self.default_creator
                log.debug("  Using default creator instead:  %s", creator)

            else:
                raise InvalidCreator(
                    f"Could not determine a plot creator for plot '{name}'!\n"
                    "Either specify it directly via the `creator` argument, "
                    "associate one with the plot function via the decorator, "
                    "or set the `default_creator` argument during the "
                    "initialization of the PlotManager."
                )

        # Determine the actual creator *type* now.
        # More precisely, this can also be a factory that then generates the
        # BasePlotCreator-like object.
        if not callable(creator):
            PlotCreator = self.CREATORS[creator]
        else:
            PlotCreator = creator

        # Parse initialization kwargs, based on the defaults set in __init__
        # FIXME This is not working properly if ``creator`` is not a string!
        pc_kwargs = recursive_update(
            copy.deepcopy(self._shared_cckwargs),
            self._cckwargs.get(creator, {}),
        )
        if init_kwargs:
            log.debug("Recursively updating creator initialization kwargs ...")
            pc_kwargs = recursive_update(copy.deepcopy(pc_kwargs), init_kwargs)

        if "raise_exc" not in pc_kwargs:
            pc_kwargs["raise_exc"] = self.raise_exc

        # Can now instantiate the creator object
        pc = PlotCreator(
            name=name, plot_func=plot_func, dm=self._dm, **pc_kwargs
        )

        log.debug("Initialized %s.", pc.logstr)
        return pc

    def _invoke_plot_creation(
        self,
        plot_creator: BasePlotCreator,
        *,
        out_path: str,
        debug: bool = None,
        **plot_cfg,
    ) -> Union[bool, str]:
        """This method wraps the plot creator's ``__call__`` and is the last
        PlotManager method that is called prior to handing over to the selected
        plot creator. It takes care of invoking the plot creator's ``__call__``
        method and handling potential error messages and return values.

        Args:
            plot_creator (BasePlotCreator): The currently used creator object
            out_path (str): The plot output path
            debug (bool, optional): If given, this overwrites the ``raise_exc``
                option specified during initialization.
            **plot_cfg: The plot configuration

        Returns:
            Union[bool, str]: Whether the plot was carried out successfully.
                Returns the string ``'skipped'`` if the plot was skipped via a
                :py:class:`~dantro.exceptions.SkipPlot` exception.

        Raises:
            PlotCreatorError: On error within the plot creator. This is only
                raised if either ``debug is True`` or
                ``debug is None and self.raise_exc``. Otherwise, the error
                message is merely logged.
        """
        try:
            plot_creator(out_path=out_path, **plot_cfg)

        except SkipPlot as skip_reason:
            log.caution("Skipped. %s\n", skip_reason)
            return "skipped"

        except DantroMessagingException as msg:
            # Pass on to outside scope
            raise

        except Exception as exc:
            self._handle_exception(
                exc, pc=plot_creator, debug=debug, ExcCls=PlotCreatorError
            )
            return False

        log.debug("Plot creator call returned successfully.")
        return True

    def _invoke_parallel_plot_creation(
        self,
        *,
        task_key: Any,
        plot_creator: BasePlotCreator,
        out_path: str,
        cfg: dict,
        plot_cfg: dict,
    ) -> Tuple[int, Tuple[str, str], Union[bool, str]]:
        """Shallow wrapper around plot creator invocation, preparing arguments
        for invocation in the context of parallel execution."""
        import contextlib
        import io

        from .logging import DantroLogger

        captured = io.StringIO("")
        DantroLogger.change_settings(
            divert_to=captured,
            suppress_in_child_process=True,
        )

        with contextlib.redirect_stdout(captured):
            rv = self._invoke_plot_creation(
                plot_creator, out_path=out_path, **cfg, **plot_cfg
            )
        return task_key, captured.getvalue(), rv

    def _invoke_parallel_executor_benchmark(self) -> bool:
        """A method that is passed to a parallel executor for benchmarking
        how long it takes to spin up a new thread or process.
        This is especially relevant for processes, because they require that
        the PlotManager and all accompanying objects are pickled and passed
        to the child process, which can take a long time ...

        The method itself does not perform any actions."""
        return True

    def _parse_parallel_plotting_kwargs(
        self,
        *,
        enabled: bool,
        executor: Literal["thread", "process"] = "process",
        max_workers: Union[int, float] = None,
        fallback_on_fail: bool = False,
        benchmark_overhead: Union[bool, int] = 5,
        **executor_kwargs,
    ) -> dict:
        """Prepares arguments for parallel plotting"""
        if isinstance(max_workers, float) and (-1.0 < max_workers <= 1.0):
            max_workers = max_workers * os.cpu_count()

        if max_workers is not None:
            if max_workers < 0:
                max_workers = os.cpu_count() + max_workers

            if max_workers <= 1:
                enabled = False

            max_workers: int = round(max_workers)

        # TODO Automatic mode? :thinking: Would need to benchmark a single plot

        return dict(
            enabled=bool(enabled),
            fallback_on_fail=fallback_on_fail,
            benchmark_overhead=benchmark_overhead,
            executor_name=executor,
            max_workers=max_workers,
            **executor_kwargs,
        )

    def _parallel_executor(
        self, executor_name: str, *, max_workers: int = None, **executor_kwargs
    ) -> "concurrent.futures.Executor":
        import concurrent.futures as concfu
        import multiprocessing as mp

        if executor_name == "process":
            ExecutorCls = concfu.ProcessPoolExecutor
            executor_kwargs["mp_context"] = mp.get_context(
                executor_kwargs.get("mp_context", "fork")
            )

        elif executor_name == "thread":
            ExecutorCls = concfu.ThreadPoolExecutor

        else:
            raise ValueError(
                f"Executor name '{executor_name}' invalid, "
                "can only be 'process' or 'thread'."
            )

        # Set up environment variables
        log.remark(
            "Setting up %s (max_workers: %s) ...",
            ExecutorCls.__name__,
            max_workers,
        )
        return ExecutorCls(max_workers=max_workers, **executor_kwargs)

    def _handle_UpdatePlotConfig(
        self, upc: UpdatePlotConfig, plot_cfg: dict
    ) -> dict:
        """Updates the plot configuration with the given updates."""
        log.note("Received plot configuration updates.")
        log.debug("Context:  %s", upc.context)
        return recursive_update(copy.deepcopy(plot_cfg), upc.plot_cfg_updates)

    def _store_plot_info(
        self,
        name: str,
        *,
        plot_cfg: dict,
        plot_cfg_extras: dict,
        creator_name: str,
        save: bool,
        target_dir: str,
        part_of_sweep: bool = False,
        **info,
    ):
        """Stores all plot information in the plot_info list and, if ``save``
        is set, also saves it using :py:meth:`._save_plot_cfg`.
        """
        # Prepare the entry
        entry = dict(
            name=name,
            plot_cfg=plot_cfg,
            plot_cfg_extras=plot_cfg_extras,
            target_dir=target_dir,
            creator_name=creator_name,
            part_of_sweep=part_of_sweep,
            **info,
            plot_cfg_path=None,
        )

        if save:
            # Save the plot configuration
            save_path = self._save_plot_cfg(
                plot_cfg,
                name=name,
                target_dir=target_dir,
                **plot_cfg_extras,
            )

            # Store the path the configuration was saved at
            entry["plot_cfg_path"] = save_path

        # Append to the plot_info list
        self._plot_info.append(entry)

    def _save_plot_cfg(
        self,
        cfg: dict,
        *,
        name: str,
        target_dir: str,
        exists_action: str = None,
        is_sweep: bool = False,
        **plot_cfg_extras,
    ) -> str:
        """Saves the given configuration under the top-level entry ``name`` to
        a yaml file.

        Args:
            cfg (dict): The plot configuration to save
            name (str): The name of the plot
            target_dir (str): The directory path to store the file in
            exists_action (str, optional): What to do if a plot configuration
                already exists. Can be: ``overwrite``, ``overwrite_nowarn``,
                ``skip``, ``append``, ``raise``. If None, uses the value of the
                ``cfg_exists_action`` argument given during initialization.
            is_sweep (bool, optional): Set if the configuration refers to a
                plot in sweep mode, for which a different format string is used
            **plot_cfg_extras: Added to the plot configuration via recursive
                update.

        Returns:
            str: The path the config was saved at (mainly used for testing)

        Raises:
            ValueError: For invalid ``exists_action`` argument
        """
        # Resolve default arguments
        if exists_action is None:
            exists_action = self._cfg_exists_action

        # Build the dict that is to be saved
        d = dict()
        d[name] = copy.deepcopy(cfg)

        # Need to include some extra information like the creator and the
        # plot function that was used
        if not isinstance(cfg, ParamSpace):
            d[name] = recursive_update(d[name], plot_cfg_extras)

        else:
            # FIXME hacky, should not use the internal API!
            d[name]._dict = recursive_update(d[name]._dict, plot_cfg_extras)

        # Generate the filename and save path
        fn_fstr = self.out_fstrs["plot_cfg_sweep" if is_sweep else "plot_cfg"]
        fname = fn_fstr.format(name=name, basename=os.path.basename(name))
        save_path = os.path.join(target_dir, fname)

        # Try to write
        try:
            write_yml(d, path=save_path, mode="x")

        except FileExistsError:
            log.debug("Config file already exists at %s!", save_path)

            if exists_action == "raise":
                raise

            elif exists_action == "skip":
                log.debug("Skipping ...")

            elif exists_action == "append":
                log.debug("Appending ...")
                write_yml(d, path=save_path, mode="a")

            elif exists_action == "overwrite":
                log.warning("Overwriting existing plot configuration ...")
                write_yml(d, path=save_path, mode="w")

            elif exists_action == "overwrite_nowarn":
                log.debug("Overwriting ...")
                write_yml(d, path=save_path, mode="w")

            else:
                raise ValueError(
                    f"Invalid value '{exists_action}' for argument "
                    "`exists_action`! Choose from: raise, skip, append, "
                    "overwrite, overwrite_nowarn"
                )

        else:
            log.debug(
                "Saved plot configuration for '%s' to: %s", name, save_path
            )

        return save_path

    # .. Plotting .............................................................

    def plot_from_cfg(
        self,
        *,
        plots_cfg: Union[dict, str] = None,
        plot_only: List[str] = None,
        out_dir: str = None,
        resolve_based_on: bool = True,
        **update_plots_cfg,
    ) -> None:
        """Create multiple plots from a configuration, either a given one or
        the one passed during initialization.

        This is mostly a wrapper around the plot function, allowing additional
        ways of how to configure and create plots.

        Args:
            plots_cfg (Union[dict, str], optional): The plots configuration to
                use. If not given, the ``default_plots_cfg`` specified during
                initialization is used. If a string is given, will assume it
                is a path and load the file.
            plot_only (List[str], optional): If given, create only those plots
                from the resulting configuration that match these names. This
                will lead to the `enabled` key being ignored, regardless of its
                value. The strings given here may also include Unix shell-like
                wildcards like ``*`` and ``? ``, which are matched using the
                Python ``fnmatch`` module.
            out_dir (str, optional): A different output directory; will use the
                one passed at initialization if the given argument evaluates to
                False.
            resolve_based_on (bool, optional): Whether to resolve the
                ``based_on`` entries in ``plots_cfg`` here. If false, will
                postpone this to :py:meth:`.plot`,
                thus *not* including the rest of the ``plots_cfg`` in the base
                configuration pool for name resolution.
                Lookups happen from ``base_cfg_pools`` following the rules
                described in :py:func:`~dantro.plot._cfg.resolve_based_on`.
            **update_plots_cfg: If given, it is used to update the plots_cfg
                recursively. Note that on the top level the _names_ of the
                plots are placed; this cannot be used to make all plots have a
                common property. Furthermore, this update happens *before* the
                ``based_on`` entries are resolved.

        Raises:
            PlotConfigError: Empty or invalid plot configuration
            ValueError: Bad ``plot_only`` argument, e.g. not matching any of
                the available plot names.
        """
        # Determine which plot configuration to use, ensuring a deep copy
        if not plots_cfg:
            if not self._default_plots_cfg and not update_plots_cfg:
                e_msg = (
                    "Got empty `plots_cfg` and `plots_cfg` given at "
                    "initialization was also empty. Nothing to plot."
                )

                if self.raise_exc:
                    raise PlotConfigError(e_msg)

                log.error(e_msg)
                return

            log.debug("Using default plots configuration.")
            plots_cfg = copy.deepcopy(self._default_plots_cfg)

        else:
            plots_cfg = self._prepare_cfg(plots_cfg)

        # Allow update of plots configuration
        if update_plots_cfg:
            plots_cfg = recursive_update(
                plots_cfg, copy.deepcopy(update_plots_cfg)
            )
            log.debug("Updated the plots configuration.")

        # Check the plot configuration for invalid types
        for plot_name, cfg in plots_cfg.items():
            if not isinstance(cfg, (dict, ParamSpace)):
                raise PlotConfigError(
                    "Got invalid plots specifications for the plot named "
                    f"'{plot_name}'! Expected dict or ParamSpace, but got "
                    f"{type(cfg).__name__} with value '{cfg}'."
                )

        # Evaluate `plot_only`, but retain a copy of the full configuration
        full_plots_cfg = copy.deepcopy(plots_cfg)
        if plot_only is None:
            # Resolve all `enabled` entries, creating a new plots_cfg dict
            plots_cfg = {
                k: v for k, v in plots_cfg.items() if v.pop("enabled", True)
            }
        else:
            # Filter the plot selection, resolving glob-like patterns
            to_plot = []
            for name in plot_only:
                if any([s in name for s in ("*", "?", "[", "]")]):
                    # Add all matching plot names
                    to_plot += fnmatch.filter(plots_cfg.keys(), name)
                else:
                    # No globbing; add the name directly
                    to_plot.append(name)

            # Reduce the plot configuration to those entries
            try:
                plots_cfg = {k: plots_cfg[k] for k in to_plot}

            except KeyError as err:
                _plot_only = ", ".join(plot_only)
                raise ValueError(
                    f"Could not find a configuration for a plot named {err} "
                    "while resolving the plot_only argument! Check that it "
                    "was specified correctly:\n"
                    f"  plot_only:  {_plot_only}\n"
                    f"  available:\n{make_columns(plots_cfg.keys())}"
                ) from err

            # Remove all `enabled` keys from the remaining entries, thus also
            # enabling those plots that were set `enabled: False`
            for cfg in plots_cfg.values():
                cfg.pop("enabled", None)

        # Also throw out entries that start with an underscore or dot
        plots_cfg = {
            k: v
            for k, v in plots_cfg.items()
            if not (k.startswith("_") or k.startswith("."))
        }

        # Now resolve the `based_on` entries of all remaining plot entries,
        # but allow lookup from those plots that were not enabled.
        if resolve_based_on:
            # Determine only the _unused_ plot configurations such that they
            # can be added as a separate pool. As these might still contain
            # the already-evaluated `enabled` key, need to pop that out.
            # NOTE The full_plots_cfg no longer contains those entries after
            #      this operation was carried out.
            unused_plots_cfg = {
                k: (v, v.pop("enabled", None))[0]
                for k, v in full_plots_cfg.items()
                if k not in plots_cfg
            }

            plots_cfg = _resolve_based_on(
                plots_cfg,
                label="plot_from_cfg",
                base_pools=(
                    tuple(self.base_cfg_pools.items())
                    + (("plot_from_cfg_unused", unused_plots_cfg),)
                ),
            )

        # Determine the output directory path to use; not creating directories!
        if not out_dir:
            out_dir = self._out_dir
        out_dir = self._parse_out_dir(out_dir, name="{name:}")
        # NOTE creating this here such that all plots from this config are side
        #      by side in one output directory. With the given `name` key, the
        #      evaluation of that part of the out_dir path is postponed
        #      to when the actual plot with that name is created.
        #      Not doing this would lead to multiple output paths for different
        #      time stamps.

        # Provide information on how many plots will be created
        log.hilight(
            "Performing plots from %d plot configuration entr%s:",
            len(plots_cfg),
            "ies" if len(plots_cfg) != 1 else "y",
        )
        t0 = time.time()

        _plot_names = []
        _num_plots = 0
        for plot_name, cfg in plots_cfg.items():
            _n = 1
            _creator = cfg.get("creator", "auto")
            if isinstance(cfg, ParamSpace):
                _n = cfg.volume if cfg.volume else 1

            _plot_names.append(
                "  - {:<50s}  ({:s}, {:d} plot{:s})"
                "".format(plot_name, _creator, _n, "s" if _n != 1 else "")
            )
            _num_plots += _n
        log.note(
            "Have (at least) the following %d plots to perform:\n%s\n",
            _num_plots,
            "\n".join(_plot_names),
        )

        # Loop over the configured plots and invoke the individual plot calls
        for plot_name, cfg in plots_cfg.items():
            if isinstance(cfg, ParamSpace):
                self.plot(plot_name, default_out_dir=out_dir, from_pspace=cfg)

            else:
                self.plot(plot_name, default_out_dir=out_dir, **cfg)

        log.success(
            "Performed plots from %d plot configuration%s in %s.\n",
            len(plots_cfg),
            "s" if len(plots_cfg) != 1 else "",
            _fmt_time(time.time() - t0),
        )

    def plot(
        self,
        name: str,
        *,
        based_on: Union[str, Tuple[str]] = None,
        from_pspace: Union[dict, ParamSpace] = None,
        **plot_cfg,
    ) -> BasePlotCreator:
        """Create plot(s) from a single configuration entry.

        A call to this function resolves the ``based_on`` feature and passes
        the derived plot configuration to :py:meth:`._plot`, which actually
        carries out the plotting. See there for documentation of further
        arguments.

        Note that more than one plot can result from a single configuration
        entry, e.g. when plots were configured that have more dimensions than
        representable in a single file.

        For

        Args:
            name (str): The name of this plot. This will be used for generating
                an output file path later on. Some characters are not allowed,
                e.g. ``*`` and ``?``, but a ``/`` can be used to store the plot
                output in a subdirectory.
            based_on (Union[str, Tuple[str]], optional): A key or a sequence
                of keys of entries in the base pool that should be used as
                the basis of this plot. The given plot configuration is then
                used to recursively update (a copy of) those base
                configuration entries.
                Lookups happen from ``base_cfg_pools`` following the rules
                described in :py:func:`~dantro.plot._cfg.resolve_based_on`.
            from_pspace (Union[dict, paramspace.paramspace.ParamSpace], optional):
                If given, execute a parameter sweep over these parameters,
                re-using the same creator instance. If this is a dict, a
                ParamSpace is created from it.
            **plot_cfg: The plot configuration, including some parameters that
                the plot manager will evaluate (and consequently: does not
                pass on to the plot creator).
                If using ``from_pspace``, parameters given here will
                recursively update those given in ``from_pspace``.

        Returns:
            BasePlotCreator: The PlotCreator used for these plots
        """
        self._check_plot_name(name)

        # Distinguish cases that are using a ParamSpace and those that do not.
        # The latter are far simpler to handle ...
        if from_pspace is None:
            plot_cfg = _resolve_based_on_single(
                name=name,
                based_on=based_on,
                plot_cfg=plot_cfg,
                label="plot",
                base_pools=self._base_cfg_pools,
            )
            try:
                return self._plot(name, **plot_cfg)

            except UpdatePlotConfig as upc:
                updated_plot_cfg = self._handle_UpdatePlotConfig(upc, plot_cfg)
                log.info("Restarting plot with updated configuration ...\n")
                return self.plot(name, **updated_plot_cfg)

        # Else: It's more complicated now, as the config is in from_pspace, and
        # (partly) in plot_cfg. Urgh.
        # To make it easier, get information out of the `ParamSpace` object,
        # move entries from `plot_cfg` over there, resolve the `based_on`
        # entries and then build a new `ParamSpace` object (done in `_plot`).
        if isinstance(from_pspace, ParamSpace):
            from_pspace = copy.deepcopy(from_pspace._dict)
            # FIXME Should not have to use private API!

        # Combine from_pspace and plot_cfg into one dict
        from_pspace = recursive_update(from_pspace, copy.deepcopy(plot_cfg))

        # Resolve `based_on`
        from_pspace = _resolve_based_on_single(
            name=name,
            based_on=based_on,
            plot_cfg=from_pspace,
            label="plot_pspace",
            base_pools=self._base_cfg_pools,
        )

        # Extract keys which `_plot` expects as separate arguments
        kwargs = {
            k: from_pspace.pop(k, None)
            for k in (
                "plot_func",
                "module",
                "module_file",
                "creator",
                "out_dir",
                "default_out_dir",
                "save_plot_cfg",
                "creator_init_kwargs",
                "parallel",
            )
        }

        try:
            return self._plot(name, from_pspace=from_pspace, **kwargs)

        except UpdatePlotConfig as upc:
            updated_plot_cfg = self._handle_UpdatePlotConfig(
                upc, dict(from_pspace=from_pspace, **kwargs)
            )
            log.note("Restarting plot with updated configuration ...\n")
            return self.plot(name, **updated_plot_cfg)

    def _plot(
        self,
        name: str,
        *,
        plot_func: Union[str, Callable] = None,
        module: str = None,
        module_file: str = None,
        creator: Union[str, Callable] = None,
        out_dir: str = None,
        default_out_dir: str = None,
        file_ext: str = None,
        save_plot_cfg: bool = None,
        creator_init_kwargs: dict = None,
        from_pspace: dict = None,
        parallel: Union[dict, bool] = None,
        **plot_cfg,
    ) -> BasePlotCreator:
        """Create plot(s) from a single configuration entry.

        This first resolves the plot function using the plot function resolver
        class: :py:class:`~dantro.plot.utils.plot_func.PlotFuncResolver` or a
        derived class (depending on the :py:attr:`.PLOT_FUNC_RESOLVER`).

        A call to this function creates a :ref:`plot creator <plot_creators>`,
        which is also returned after all plots are finished.

        Note that more than one plot can result from a single configuration
        entry, e.g. when plots were configured that have more dimensions than
        representable in a single file or when using ``from_pspace``.

        Args:
            name (str): The name of this plot
            plot_func (Union[str, Callable], optional): The name or module
                string of the plot function as it can be imported from
                ``module``. If this is a callable will directly return that
                callable. This argument *needs* be given.
            module (str): If ``plot_func`` was the name of the plot
                function, this needs to be the name of the module to import
                that name from.
            module_file (str): Path to the file to load and look for
                the ``plot_func`` in. If ``base_module_file_dir`` is given
                during initialization, this can also be a path relative to that
                directory.
            creator (Union[str, Callable]): The name of the creator to
                be looked up in :py:attr:`.CREATORS`. Can also be None, in
                which case it is attempted to look it up from the ``plot_func``
                's ``creator`` attribute. If that was not possible either, the
                :py:attr:`.default_creator` is used. If a callable is given,
                will use that as a factory to set up the creator.
            out_dir (str, optional): If given, will use this directory as out
                directory. If not, will use the default value given by
                ``default_out_dir`` or that given at initialization.
            default_out_dir (str, optional): An output directory that was
                determined *in the calling context* and which should be used as
                default if no ``out_dir`` was given explicitly.
            file_ext (str, optional): The file extension to use, including the
                leading dot!
            save_plot_cfg (bool, optional): Whether to save the plot config.
                If not given, uses the default value from initialization.
            creator_init_kwargs (dict, optional): Passed to the plot creator
                during initialization. Note that the arguments given at
                initialization of the PlotManager are updated by this.
            from_pspace (dict, optional): If given, execute a parameter
                sweep over this parameter space, re-using the same creator
                instance. Each point in parameter space will end up calling
                this method with arguments unpacked to the ``plot_cfg``
                argument.
            **plot_cfg: The plot configuration to pass on to the plot creator.
                This may be completely empty if ``from_pspace`` is used!

        Returns:
            BasePlotCreator: The PlotCreator used for these plots. This will
                also be returned in case the plot failed!

        Raises:
            PlotConfigError: If no out directory was specified here or at
                initialization.
            PlotCreatorError: In case the preparation or execution of the plot
                failed for whatever reason. Not raised if not in debug mode.
        """

        log.debug("Preparing plot '%s' ...", name)
        t0 = time.time()

        # Gather arguments that are explicitly handled here; they may still be
        # needed downstream, e.g. when storing the plot configuration to file.
        plot_cfg_extras = dict(
            out_dir=out_dir,
            file_ext=file_ext,
            creator=creator,
            plot_func=str(plot_func),  # ... to not have a callable in here
            module=module,
            module_file=module_file,
            creator_init_kwargs=copy.deepcopy(creator_init_kwargs),
            save_plot_cfg=save_plot_cfg,
            parallel=parallel,
        )
        plot_cfg_extras = {
            k: v for k, v in plot_cfg_extras.items() if v is not None
        }

        # Evaluate the output directory and whether to save plot configs
        if not out_dir:
            if default_out_dir:
                out_dir = default_out_dir

            elif self._out_dir:
                out_dir = self._out_dir

            else:
                raise PlotConfigError(
                    "No `out_dir` specified here and at "
                    "initialization; cannot perform plot."
                )

        if save_plot_cfg is None:
            save_plot_cfg = self.save_plot_cfg

        # Retrieve the plot function callable
        if plot_func is None:
            raise PlotConfigError("Missing `plot_func` argument!")
        plot_func = self._get_plot_func(
            plot_func=plot_func, module=module, module_file=module_file
        )

        # Set up the plot creator instance
        plot_creator = self._get_plot_creator(
            creator=creator,
            plot_func=plot_func,
            name=name,
            init_kwargs=creator_init_kwargs,
        )

        # Let the creator process arguments
        try:
            plot_cfg, from_pspace = plot_creator.prepare_cfg(
                plot_cfg=plot_cfg, pspace=from_pspace
            )
        except Exception as exc:
            _debug = plot_cfg.get("debug")
            self._handle_exception(
                exc, pc=plot_creator, debug=_debug, ExcCls=PlotCreatorError
            )
            return plot_creator

        # Distinguish single calls and parameter sweeps
        if not from_pspace:
            log.progress("Plotting '%s' ...", name)

            # Generate the output path
            out_dir = self._parse_out_dir(out_dir, name=name)
            out_path = self._parse_out_path(
                plot_creator, name=name, out_dir=out_dir, file_ext=file_ext
            )

            # Call the plot creator to perform the plot, using the private
            # method to perform exception handling
            rv = self._invoke_plot_creation(
                plot_creator, out_path=out_path, **plot_cfg
            )

            # Store plot information (_save_ only if not skipped)
            self._store_plot_info(
                name=name,
                creator_name=creator,
                out_path=out_path,
                plot_cfg=plot_cfg,
                plot_cfg_extras=plot_cfg_extras,
                save=save_plot_cfg and rv != "skipped",
                target_dir=os.path.dirname(out_path),
                creator_rv=rv,
                part_of_sweep=False,
            )

            if rv is True:
                log.progress(
                    "Performed '%s' plot in %s.\n\n",
                    name,
                    _fmt_time(time.time() - t0),
                )

            return plot_creator

        # else: Is a parameter sweep over the plot configuration ..............
        # NOTE The parameter space is allowed to have volume 0!

        # Make sure it's a ParamSpace
        if not isinstance(from_pspace, ParamSpace):
            from_pspace = ParamSpace(from_pspace)

        # Extract some info and communicate it
        psp_vol = from_pspace.volume
        psp_dims = from_pspace.dims

        if psp_vol > 0:
            amap_coords = from_pspace.active_state_map.coords
            max_dname_len = max(len(n) for n in amap_coords.keys())
            n_max = psp_vol

            log.progress("Performing %d '%s' plots ...", n_max, name)
            log.note(
                "... iterating over parameter space:\n%s",
                "\n".join(
                    [
                        "  * {0:<{d:}} : {1:}".format(
                            dim_name,
                            ", ".join([str(c) for c in coords.values]),
                            d=max_dname_len,
                        )
                        for dim_name, coords in amap_coords.items()
                    ]
                ),
            )

        else:
            log.progress("Plotting '%s' ...", name)
            log.note("... from default point in zero-volume parameter space.")
            n_max = 1

        # Parse the output directory, such that all plots are together in
        # one directory even if the timestamp varies
        out_dir = self._parse_out_dir(out_dir, name=name)

        # Parallel execution?
        parallel = self._parse_parallel_plotting_kwargs(
            **(
                parallel
                if isinstance(parallel, dict)
                else dict(enabled=parallel)
            )
        )

        # Evaluate whether to do this in sequence or in parallel.
        # There is also the option to fall back to the sequential case, if
        # there was an error in parallel execution.
        plot_sequential = not parallel.pop("enabled", False)
        fallback_on_fail = parallel.pop("fallback_on_fail", False)
        if not plot_sequential and n_max > 1:
            try:
                res = self._plot_pspace_parallel(
                    from_pspace,
                    name=name,
                    creator_name=creator,
                    plot_creator=plot_creator,
                    out_dir=out_dir,
                    file_ext=file_ext,
                    plot_cfg=plot_cfg,
                    plot_cfg_extras=plot_cfg_extras,
                    save_plot_cfg=save_plot_cfg,
                    t0=t0,
                    psp_vol=psp_vol,
                    psp_dims=psp_dims,
                    n_max=n_max,
                    **parallel,
                )
            except Exception as exc:
                raise
                log.error(
                    "Parallel parameter space plot of '%s' failed!\n"
                    "Got %s: %s\n",
                    name,
                    type(exc).__name__,
                    exc,
                )
                if not fallback_on_fail:
                    log.caution(
                        "Check if the error was caused by parallel plotting "
                        "or by the plot itself, regardless of execution.\n"
                        "To test this, disable parallel plotting or set the "
                        "`fallback_on_fail` flag to re-try with sequential "
                        "execution."
                    )
                    raise
                log.caution("Re-trying using non-parallel plot ...")
                plot_sequential = True

        if plot_sequential:
            res = self._plot_pspace(
                from_pspace,
                name=name,
                creator_name=creator,
                plot_creator=plot_creator,
                out_dir=out_dir,
                file_ext=file_ext,
                plot_cfg=plot_cfg,
                plot_cfg_extras=plot_cfg_extras,
                save_plot_cfg=save_plot_cfg,
                t0=t0,
                psp_vol=psp_vol,
                psp_dims=psp_dims,
                n_max=n_max,
            )

        # Finished parameter space iteration.
        num_skipped = res["num_skipped"]

        # Save the plot configuration alongside, if configured to do so, and
        # if at least one of the plots was *not* skipped and the parameter
        # space was not zero-volume
        if save_plot_cfg and num_skipped < n_max and psp_vol > 0:
            self._save_plot_cfg(
                from_pspace,
                name=name,
                target_dir=out_dir,
                is_sweep=True,
                **plot_cfg_extras,
            )

        dt = time.time() - t0
        if not num_skipped:
            log.progress(
                "Performed all %d '%s' plots in %s.",
                n_max,
                name,
                _fmt_time(dt),
            )
        else:
            log.progress(
                "Performed %d/%d '%s' plots in %s, skipped %d.",
                n_max - num_skipped,
                n_max,
                name,
                _fmt_time(dt),
                num_skipped,
            )
        if n_max > 1 and num_skipped < n_max:
            log.remark(
                "Average:  %s / plot\n\n",
                _fmt_time(dt / (n_max - num_skipped)),
            )
        else:
            log.remark("\n")

        # Done now. Return the plot creator object
        return plot_creator

    def _plot_pspace(
        self,
        from_pspace: ParamSpace,
        *,
        name: str,
        creator_name: str,
        plot_creator: BasePlotCreator,
        out_dir: str,
        file_ext: str,
        plot_cfg: dict,
        plot_cfg_extras: dict,
        save_plot_cfg: bool,
        t0: float,
        psp_vol: int,
        psp_dims: list,
        n_max: int,
    ) -> dict:
        """Performs parameter sweep plots in sequence."""
        # Keep track of how many plots are skipped
        num_skipped = 0

        # Create the iterator
        it = from_pspace.iterator(
            with_info=("state_no", "state_vector", "coords")
        )

        # ...and loop over all points:
        for n, (cfg, state_no, state_vector, coords) in enumerate(it):
            log.progress("Plotting '%s' (%d/%d) ...", name, n + 1, n_max)
            if coords:
                log.note(
                    "Current coordinates of sweep plot configuration:\n%s",
                    "\n".join(
                        "  {:>23s}:   {}".format(*kv) for kv in coords.items()
                    ),
                )

            # Handle the file extension parameter; it might come from the
            # given configuration and then needs to be popped such that it
            # is not propagated to the plot creator.
            _file_ext = cfg.pop("file_ext", file_ext)

            # Generate the output path
            out_path = self._parse_out_path(
                plot_creator,
                name=name,
                out_dir=out_dir,
                file_ext=_file_ext,
                state_no=state_no,
                state_no_max=n_max - 1,
                state_vector=state_vector,
                dims=psp_dims,
            )

            # Call the plot creator to perform the plot, using the private
            # method to perform exception handling
            rv = self._invoke_plot_creation(
                plot_creator, out_path=out_path, **cfg, **plot_cfg
            )
            # NOTE The **plot_cfg is passed here in order to not lose any
            #      arguments that might have been passed to it. While `cfg`
            #      _should_ hold all the arguments from the parameter space
            #      iteration, there might be more arguments in `plot_cfg`;
            #      rather than disallowing this, we pass them on and
            #      forward responsibility downstream ...

            # Always store plot information, regardless of skipping.
            # Saving is enabled only for zero-volume parameter sweeps in order
            # to have the backup file right beside the plot; the file will not
            # be saved again after the for-loop.
            self._store_plot_info(
                name=name,
                creator_name=creator_name,
                out_path=out_path,
                plot_cfg=dict(**cfg, **plot_cfg),
                plot_cfg_extras=plot_cfg_extras,
                state_no=state_no,
                state_vector=state_vector,
                save=(save_plot_cfg and psp_vol == 0 and rv != "skipped"),
                target_dir=os.path.dirname(out_path),
                creator_rv=rv,
                part_of_sweep=True,
            )

            # Count skipped plots
            if rv == "skipped":
                num_skipped += 1

            elif rv is True:
                log.progress("Finished plot %d/%d.", n + 1, n_max)

                # Estimate for time remaining
                if (n + 1) < n_max:
                    dt = time.time() - t0
                    etl = dt / ((n + 1) / n_max) - dt
                    log.note(
                        "Estimated time needed for %d remaining plots:  %s\n",
                        n_max - (n + 1),
                        _fmt_time(etl),
                    )

        return dict(num_skipped=num_skipped)

    def _plot_pspace_parallel(
        self,
        from_pspace: ParamSpace,
        *,
        name: str,
        creator_name: str,
        plot_creator: BasePlotCreator,
        out_dir: str,
        file_ext: str,
        plot_cfg: dict,
        plot_cfg_extras: dict,
        save_plot_cfg: bool,
        t0: float,
        psp_vol: int,
        psp_dims: list,
        n_max: int,
        #
        executor_name: Literal["thread", "process"],
        benchmark_overhead: Union[int, bool],
        show_exception_summary: bool = True,
        **executor_kwargs,
    ) -> dict:
        """Performs parameter sweep plots in sequence."""
        import concurrent.futures as concfu

        log.remark("Creating plotting tasks ...")

        # .. Generate the tasks
        it = from_pspace.iterator(
            with_info=("state_no", "state_vector", "coords")
        )
        tasks = dict()
        pspace_info = dict()
        debug = None

        for n, (cfg, state_no, state_vector, coords) in enumerate(it):
            log.debug("Adding '%s' plot task (%d/%d) ...", name, n + 1, n_max)

            # Handle the file extension parameter; it might come from the
            # given configuration and then needs to be popped such that it
            # is not propagated to the plot creator.
            _file_ext = cfg.pop("file_ext", file_ext)

            # Generate the output path
            out_path = self._parse_out_path(
                plot_creator,
                name=name,
                out_dir=out_dir,
                file_ext=_file_ext,
                state_no=state_no,
                state_no_max=n_max - 1,
                state_vector=state_vector,
                dims=psp_dims,
            )

            tasks[state_no] = dict(
                plot_creator=plot_creator,
                out_path=out_path,
                cfg=cfg,
                plot_cfg=plot_cfg,
            )
            pspace_info[state_no] = dict(
                state_no=state_no, state_vector=state_vector, coords=coords
            )

            # Evaluate debug flag; one `debug: true` suffices to trigger the
            # debugging mode. If it is not set anywhere, it will remain None.
            if not debug:
                debug = cfg.get("debug", plot_cfg.get("debug"))

        # .. Parallel plotting
        futures: List[concfu.Future] = []
        exceptions: Dict[int, Exception] = dict()
        executor = self._parallel_executor(executor_name, **executor_kwargs)

        with executor:
            t0 = time.time()

            if benchmark_overhead:
                num_bfs = int(benchmark_overhead)
                log.remark(
                    "Benchmarking executor spawning overhead using %d tasks …",
                    num_bfs,
                )
                bfs = [
                    executor.submit(self._invoke_parallel_executor_benchmark)
                    for i in range(num_bfs)
                ]
                concfu.wait(bfs)

                log.note(
                    "Executor overhead:  %s",
                    _fmt_time((time.time() - t0) / len(bfs)),
                )

            log.remark(
                "Submitting %d tasks to %s ...",
                len(tasks),
                type(executor).__name__,
            )
            for task_key, task_kwargs in tasks.items():
                future = executor.submit(
                    self._invoke_parallel_plot_creation,
                    task_key=task_key,
                    **task_kwargs,
                )
                futures.append(future)

                # TODO Can add callbacks to future here perhaps?

            log.note("Parallel plotting now commencing ...")
            t1 = time.time()

            num_skipped = 0
            num_failed = 0
            for n, future in enumerate(concfu.as_completed(futures)):
                exc = future.exception()
                if exc:
                    log.error(
                        "Parallel plotting task %s failed with a %s: %s",
                        future,
                        type(exc).__name__,
                        exc,
                    )
                    num_failed += 1
                    exceptions[n] = exc
                    continue

                # Get result
                log.debug("Task completed:  %s ", future)
                task_key, captured_stdout, rv = future.result()

                # Get parameters
                task = tasks[task_key]
                out_path = task["out_path"]
                cfg = task["cfg"]
                plot_cfg = task["plot_cfg"]
                state_no = task_key
                state_vector = pspace_info[task_key]["state_vector"]
                coords = pspace_info[task_key]["coords"]

                # Always store plot information, regardless of skipping.
                # Saving is enabled only for zero-volume parameter sweeps in
                # order to have the backup file right beside the plot; the file
                # will not be saved again after the for-loop.
                self._store_plot_info(
                    name=name,
                    creator_name=creator_name,
                    out_path=out_path,
                    plot_cfg=dict(**cfg, **plot_cfg),
                    plot_cfg_extras=plot_cfg_extras,
                    state_no=state_no,
                    state_vector=state_vector,
                    save=(save_plot_cfg and psp_vol == 0 and rv != "skipped"),
                    target_dir=os.path.dirname(out_path),
                    creator_rv=rv,
                    part_of_sweep=True,
                )

                # Inform about result
                if rv == "skipped":
                    num_skipped += 1
                    log.caution("Skipped plot %d/%d.", n + 1, n_max)
                    continue

                log.progress(
                    "Getting results for plot %d/%d ...", n + 1, n_max
                )
                if coords:
                    log.note(
                        "Coordinates:\n%s",
                        "\n".join(
                            "  {:>23s}:   {}".format(*kv)
                            for kv in coords.items()
                        ),
                    )
                if captured_stdout:
                    log.remark(
                        "Captured output:\n\n%s\n",
                        textwrap.indent(captured_stdout, " " * 4),
                    )

                # Estimate for time remaining
                log.progress("Finished plot %d/%d.", n + 1, n_max)
                if (n + 1) < n_max:
                    dt = time.time() - t1
                    etl = dt / ((n + 1) / n_max) - dt
                    log.note(
                        "Estimated time needed for %d remaining plots:  %s\n\n",
                        n_max - (n + 1),
                        _fmt_time(etl),
                    )

        log.note(
            "Finished parallel plotting of %d tasks (%d skipped, %d failed).",
            len(tasks),
            num_skipped,
            num_failed,
        )

        if exceptions:
            should_raise = debug or (debug is None and self.raise_exc)
            exc = ParallelPlottingError(
                exceptions, create_summary=show_exception_summary
            )
            if should_raise:
                raise exc
            log.error(exc)

        return dict(num_skipped=num_skipped, num_failed=num_failed)



"""The groups sub-package implements :py:class:`~dantro.base.BaseDataGroup` specializations.

isort:skip_file
"""

# To avoid circular imports, need to import this first
from ..base import BaseDataGroup
from ._registry import register_group, is_group, GROUPS
from .ordered import IndexedDataGroup, OrderedDataGroup

# The groups below can be imported in whatever order
from .dirpath import DirectoryGroup
from .graph import GraphGroup
from .labelled import LabelledDataGroup
from .psp import ParamSpaceGroup, ParamSpaceStateGroup
from .time_series import HeterogeneousTimeSeriesGroup, TimeSeriesGroup
